"""
***************************************************************************
    EquivalentNumField.py
    ---------------------
    Date                 : January 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "January 2013"
__copyright__ = "(C) 2013, Victor Olaya"

import warnings

from qgis.core import (
    QgsExpression,
    QgsExpressionContext,
    QgsExpressionContextUtils,
    QgsFeatureRequest,
    QgsProcessingException,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterExpression,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterString,
    QgsProcessingUtils,
)
from qgis.PyQt.QtCore import QCoreApplication

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm
from processing.tools import vector


class VectorLayerScatterplot(QgisAlgorithm):
    INPUT = "INPUT"
    OUTPUT = "OUTPUT"
    XFIELD = "XFIELD"
    YFIELD = "YFIELD"
    HOVERTEXT = "HOVERTEXT"
    TITLE = "TITLE"
    XAXIS_TITLE = "XAXIS_TITLE"
    YAXIS_TITLE = "YAXIS_TITLE"
    XAXIS_LOG = "XAXIS_LOG"
    YAXIS_LOG = "YAXIS_LOG"

    def group(self):
        return self.tr("Plots")

    def groupId(self):
        return "plots"

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterFeatureSource(self.INPUT, self.tr("Input layer"))
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.XFIELD,
                self.tr("X attribute"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.YFIELD,
                self.tr("Y attribute"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Numeric,
            )
        )

        self.addParameter(
            QgsProcessingParameterExpression(
                self.HOVERTEXT,
                self.tr("Hover text"),
                parentLayerParameterName=self.INPUT,
                optional=True,
            )
        )

        self.addParameter(
            QgsProcessingParameterString(self.TITLE, self.tr("Title"), optional=True)
        )

        self.addParameter(
            QgsProcessingParameterString(
                self.XAXIS_TITLE, self.tr("X-axis title"), optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterString(
                self.YAXIS_TITLE, self.tr("Y-axis title"), optional=True
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.XAXIS_LOG,
                self.tr("Use logarithmic scale for x-axis"),
                defaultValue=False,
            )
        )

        self.addParameter(
            QgsProcessingParameterBoolean(
                self.YAXIS_LOG,
                self.tr("Use logarithmic scale for y-axis"),
                defaultValue=False,
            )
        )

        self.addParameter(
            QgsProcessingParameterFileDestination(
                self.OUTPUT, self.tr("Scatterplot"), self.tr("HTML files (*.html)")
            )
        )

    def name(self):
        return "vectorlayerscatterplot"

    def displayName(self):
        return self.tr("Vector layer scatterplot")

    def shortDescription(self):
        return self.tr("Creates a simple X - Y scatter plot for a vector layer.")

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates a simple X - Y scatter plot for a vector layer."
        )

    def processAlgorithm(self, parameters, context, feedback):
        try:
            # importing plotly throws Python warnings from within the library - filter these out
            with warnings.catch_warnings():
                warnings.filterwarnings("ignore", category=ResourceWarning)
                warnings.filterwarnings("ignore", category=ImportWarning)
                import plotly as plt
                import plotly.graph_objs as go
        except ImportError:
            raise QgsProcessingException(
                QCoreApplication.translate(
                    "VectorLayerScatterplot",
                    "This algorithm requires the Python “plotly” library. Please install this library and try again.",
                )
            )

        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        xfieldname = self.parameterAsString(parameters, self.XFIELD, context)
        yfieldname = self.parameterAsString(parameters, self.YFIELD, context)

        title = self.parameterAsString(parameters, self.TITLE, context)
        xaxis_title = self.parameterAsString(parameters, self.XAXIS_TITLE, context)
        yaxis_title = self.parameterAsString(parameters, self.YAXIS_TITLE, context)
        xaxis_log = self.parameterAsBool(parameters, self.XAXIS_LOG, context)
        yaxis_log = self.parameterAsBool(parameters, self.YAXIS_LOG, context)

        if title.strip() == "":
            title = None
        if xaxis_title == "":
            xaxis_title = xfieldname
        elif xaxis_title == " ":
            xaxis_title = None
        if yaxis_title == "":
            yaxis_title = yfieldname
        elif yaxis_title == " ":
            yaxis_title = None

        output = self.parameterAsFileOutput(parameters, self.OUTPUT, context)

        hoverexpression = self.parameterAsExpression(
            parameters, self.HOVERTEXT, context
        )
        if hoverexpression.strip():
            exp_context = self.createExpressionContext(parameters, context, source)
            hoverexpression = QgsExpression(hoverexpression)
            if hoverexpression.hasParserError():
                feedback.reportError(
                    f"Expression evaluation error: {hoverexpression.evalErrorString()}"
                )
                hoverexpression = None
        else:
            hoverexpression = None

        if xfieldname == yfieldname:
            fields = [xfieldname]
        else:
            fields = [xfieldname, yfieldname]

        values = {}
        hovertext = []

        for field in fields:
            values[field] = []

        request = QgsFeatureRequest().setFlags(QgsFeatureRequest.Flag.NoGeometry)
        for feature in source.getFeatures(request):
            for field in fields:
                try:
                    v = float(feature[field])
                except ValueError:
                    v = None
                values[field].append(v)
            if hoverexpression:
                exp_context.setFeature(feature)
                txt = str(hoverexpression.evaluate(exp_context))
                if hoverexpression.hasEvalError():
                    txt = ""
                    feedback.reportError(
                        f"Expression evaluation error: {hoverexpression.evalErrorString()}"
                    )
                hovertext.append(str(txt))

        data = [go.Scatter(x=values[xfieldname], y=values[yfieldname], mode="markers")]
        fig = go.Figure(
            data=data,
            layout_title_text=title,
            layout_xaxis_title=xaxis_title,
            layout_yaxis_title=yaxis_title,
        )

        if xaxis_log:
            fig.update_xaxes(type="log")

        if yaxis_log:
            fig.update_yaxes(type="log")

        if hoverexpression:
            fig.update_traces(text=hovertext)

        fig.write_html(output)

        return {self.OUTPUT: output}



"""
***************************************************************************
    PointDistance.py
    ---------------------
    Date                 : August 2012
    Copyright            : (C) 2012 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "August 2012"
__copyright__ = "(C) 2012, Victor Olaya"

import math
import os

from qgis.core import (
    Qgis,
    QgsApplication,
    QgsDistanceArea,
    QgsFeature,
    QgsFeatureRequest,
    QgsFeatureSink,
    QgsField,
    QgsFields,
    QgsGeometry,
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterField,
    QgsProcessingParameterNumber,
    QgsProject,
    QgsSpatialIndex,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QMetaType
from qgis.PyQt.QtGui import QIcon

from processing.algs.qgis.QgisAlgorithm import QgisAlgorithm

pluginPath = os.path.split(os.path.split(os.path.dirname(__file__))[0])[0]


class PointDistance(QgisAlgorithm):
    INPUT = "INPUT"
    INPUT_FIELD = "INPUT_FIELD"
    TARGET = "TARGET"
    TARGET_FIELD = "TARGET_FIELD"
    MATRIX_TYPE = "MATRIX_TYPE"
    NEAREST_POINTS = "NEAREST_POINTS"
    OUTPUT = "OUTPUT"

    def icon(self):
        return QgsApplication.getThemeIcon("/algorithms/mAlgorithmDistanceMatrix.svg")

    def svgIconPath(self):
        return QgsApplication.iconPath("/algorithms/mAlgorithmDistanceMatrix.svg")

    def group(self):
        return self.tr("Vector analysis")

    def groupId(self):
        return "vectoranalysis"

    def documentationFlags(self):
        return Qgis.ProcessingAlgorithmDocumentationFlag.RespectsEllipsoid

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.mat_types = [
            self.tr("Linear (N*k x 3) distance matrix"),
            self.tr("Standard (N x T) distance matrix"),
            self.tr("Summary distance matrix (mean, std. dev., min, max)"),
        ]

        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.INPUT,
                self.tr("Input point layer"),
                [QgsProcessing.SourceType.TypeVectorPoint],
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.INPUT_FIELD,
                self.tr("Input unique ID field"),
                parentLayerParameterName=self.INPUT,
                type=QgsProcessingParameterField.DataType.Any,
            )
        )
        self.addParameter(
            QgsProcessingParameterFeatureSource(
                self.TARGET,
                self.tr("Target point layer"),
                [QgsProcessing.SourceType.TypeVectorPoint],
            )
        )
        self.addParameter(
            QgsProcessingParameterField(
                self.TARGET_FIELD,
                self.tr("Target unique ID field"),
                parentLayerParameterName=self.TARGET,
                type=QgsProcessingParameterField.DataType.Any,
            )
        )
        self.addParameter(
            QgsProcessingParameterEnum(
                self.MATRIX_TYPE,
                self.tr("Output matrix type"),
                options=self.mat_types,
                defaultValue=0,
            )
        )
        self.addParameter(
            QgsProcessingParameterNumber(
                self.NEAREST_POINTS,
                self.tr("Use only the nearest (k) target points"),
                type=QgsProcessingParameterNumber.Type.Integer,
                minValue=0,
                defaultValue=0,
            )
        )

        self.addParameter(
            QgsProcessingParameterFeatureSink(
                self.OUTPUT,
                self.tr("Distance matrix"),
                QgsProcessing.SourceType.TypeVectorPoint,
            )
        )

    def name(self):
        return "distancematrix"

    def displayName(self):
        return self.tr("Distance matrix")

    def shortDescription(self):
        return self.tr(
            "Creates a table containing a matrix of distances between all the points in a points layer."
        )

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates a table containing a distance matrix, with distances between all the points in a points layer."
        )

    def processAlgorithm(self, parameters, context, feedback):
        source = self.parameterAsSource(parameters, self.INPUT, context)
        if source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.INPUT)
            )

        if QgsWkbTypes.isMultiType(source.wkbType()):
            raise QgsProcessingException(
                self.tr(
                    "Input point layer is a MultiPoint layer - first convert to single points before using this algorithm."
                )
            )

        source_field = self.parameterAsString(parameters, self.INPUT_FIELD, context)
        target_source = self.parameterAsSource(parameters, self.TARGET, context)
        if target_source is None:
            raise QgsProcessingException(
                self.invalidSourceError(parameters, self.TARGET)
            )

        if QgsWkbTypes.isMultiType(target_source.wkbType()):
            raise QgsProcessingException(
                self.tr(
                    "Target point layer is a MultiPoint layer - first convert to single points before using this algorithm."
                )
            )

        target_field = self.parameterAsString(parameters, self.TARGET_FIELD, context)
        same_source_and_target = parameters[self.INPUT] == parameters[self.TARGET]
        matType = self.parameterAsEnum(parameters, self.MATRIX_TYPE, context)
        nPoints = self.parameterAsInt(parameters, self.NEAREST_POINTS, context)

        if nPoints < 1:
            nPoints = target_source.featureCount()

        if matType == 0:
            # Linear distance matrix
            return self.linearMatrix(
                parameters,
                context,
                source,
                source_field,
                target_source,
                target_field,
                same_source_and_target,
                matType,
                nPoints,
                feedback,
            )
        elif matType == 1:
            # Standard distance matrix
            return self.regularMatrix(
                parameters,
                context,
                source,
                source_field,
                target_source,
                target_field,
                nPoints,
                feedback,
            )
        elif matType == 2:
            # Summary distance matrix
            return self.linearMatrix(
                parameters,
                context,
                source,
                source_field,
                target_source,
                target_field,
                same_source_and_target,
                matType,
                nPoints,
                feedback,
            )

    def linearMatrix(
        self,
        parameters,
        context,
        source,
        inField,
        target_source,
        targetField,
        same_source_and_target,
        matType,
        nPoints,
        feedback,
    ):

        if same_source_and_target:
            # need to fetch an extra point from the index, since the closest match will always be the same
            # as the input feature
            nPoints += 1

        inIdx = source.fields().lookupField(inField)
        outIdx = target_source.fields().lookupField(targetField)

        fields = QgsFields()
        input_id_field = source.fields()[inIdx]
        input_id_field.setName("InputID")
        fields.append(input_id_field)
        if matType == 0:
            target_id_field = target_source.fields()[outIdx]
            target_id_field.setName("TargetID")
            fields.append(target_id_field)
            fields.append(QgsField("Distance", QMetaType.Type.Double))
        else:
            fields.append(QgsField("MEAN", QMetaType.Type.Double))
            fields.append(QgsField("STDDEV", QMetaType.Type.Double))
            fields.append(QgsField("MIN", QMetaType.Type.Double))
            fields.append(QgsField("MAX", QMetaType.Type.Double))

        out_wkb = (
            QgsWkbTypes.multiType(source.wkbType())
            if matType == 0
            else source.wkbType()
        )
        (sink, dest_id) = self.parameterAsSink(
            parameters, self.OUTPUT, context, fields, out_wkb, source.sourceCrs()
        )
        if sink is None:
            raise QgsProcessingException(self.invalidSinkError(parameters, self.OUTPUT))

        index = QgsSpatialIndex(
            target_source.getFeatures(
                QgsFeatureRequest()
                .setSubsetOfAttributes([])
                .setDestinationCrs(source.sourceCrs(), context.transformContext())
            ),
            feedback,
        )

        distArea = QgsDistanceArea()
        distArea.setSourceCrs(source.sourceCrs(), context.transformContext())
        distArea.setEllipsoid(context.ellipsoid())

        features = source.getFeatures(
            QgsFeatureRequest().setSubsetOfAttributes([inIdx])
        )
        total = 100.0 / source.featureCount() if source.featureCount() else 0
        for current, inFeat in enumerate(features):
            if feedback.isCanceled():
                break

            inGeom = inFeat.geometry()
            inID = str(inFeat[inIdx])
            featList = index.nearestNeighbor(inGeom.asPoint(), nPoints)
            distList = []
            vari = 0.0
            request = (
                QgsFeatureRequest()
                .setFilterFids(featList)
                .setSubsetOfAttributes([outIdx])
                .setDestinationCrs(source.sourceCrs(), context.transformContext())
            )
            for outFeat in target_source.getFeatures(request):
                if feedback.isCanceled():
                    break

                if same_source_and_target and inFeat.id() == outFeat.id():
                    continue

                outID = outFeat[outIdx]
                outGeom = outFeat.geometry()
                dist = distArea.measureLine(inGeom.asPoint(), outGeom.asPoint())

                if matType == 0:
                    out_feature = QgsFeature()
                    out_geom = QgsGeometry.unaryUnion(
                        [inFeat.geometry(), outFeat.geometry()]
                    )
                    out_feature.setGeometry(out_geom)
                    out_feature.setAttributes([inID, outID, dist])
                    sink.addFeature(out_feature, QgsFeatureSink.Flag.FastInsert)
                else:
                    distList.append(float(dist))

            if matType != 0:
                mean = sum(distList) / len(distList)
                for i in distList:
                    vari += (i - mean) * (i - mean)
                vari = math.sqrt(vari / len(distList))

                out_feature = QgsFeature()
                out_feature.setGeometry(inFeat.geometry())
                out_feature.setAttributes(
                    [inID, mean, vari, min(distList), max(distList)]
                )
                sink.addFeature(out_feature, QgsFeatureSink.Flag.FastInsert)

            feedback.setProgress(int(current * total))

        sink.finalize()
        return {self.OUTPUT: dest_id}

    def regularMatrix(
        self,
        parameters,
        context,
        source,
        inField,
        target_source,
        targetField,
        nPoints,
        feedback,
    ):

        distArea = QgsDistanceArea()
        distArea.setSourceCrs(source.sourceCrs(), context.transformContext())
        distArea.setEllipsoid(context.ellipsoid())

        inIdx = source.fields().lookupField(inField)
        targetIdx = target_source.fields().lookupField(targetField)

        index = QgsSpatialIndex(
            target_source.getFeatures(
                QgsFeatureRequest()
                .setSubsetOfAttributes([])
                .setDestinationCrs(source.sourceCrs(), context.transformContext())
            ),
            feedback,
        )

        first = True
        sink = None
        dest_id = None
        features = source.getFeatures(
            QgsFeatureRequest().setSubsetOfAttributes([inIdx])
        )
        total = 100.0 / source.featureCount() if source.featureCount() else 0
        for current, inFeat in enumerate(features):
            if feedback.isCanceled():
                break

            inGeom = inFeat.geometry()
            if first:
                featList = index.nearestNeighbor(inGeom.asPoint(), nPoints)
                first = False
                fields = QgsFields()
                input_id_field = source.fields()[inIdx]
                input_id_field.setName("ID")
                fields.append(input_id_field)
                for f in target_source.getFeatures(
                    QgsFeatureRequest()
                    .setFilterFids(featList)
                    .setSubsetOfAttributes([targetIdx])
                    .setDestinationCrs(source.sourceCrs(), context.transformContext())
                ):
                    fields.append(QgsField(str(f[targetField]), QMetaType.Type.Double))

                (sink, dest_id) = self.parameterAsSink(
                    parameters,
                    self.OUTPUT,
                    context,
                    fields,
                    source.wkbType(),
                    source.sourceCrs(),
                )
                if sink is None:
                    raise QgsProcessingException(
                        self.invalidSinkError(parameters, self.OUTPUT)
                    )

            data = [inFeat[inField]]
            for target in target_source.getFeatures(
                QgsFeatureRequest()
                .setSubsetOfAttributes([])
                .setFilterFids(featList)
                .setDestinationCrs(source.sourceCrs(), context.transformContext())
            ):
                if feedback.isCanceled():
                    break
                outGeom = target.geometry()
                dist = distArea.measureLine(inGeom.asPoint(), outGeom.asPoint())
                data.append(dist)

            out_feature = QgsFeature()
            out_feature.setGeometry(inGeom)
            out_feature.setAttributes(data)
            sink.addFeature(out_feature, QgsFeatureSink.Flag.FastInsert)
            feedback.setProgress(int(current * total))

        return {self.OUTPUT: dest_id}



"""
/***************************************************************************
Python Console for QGIS
                             -------------------
begin                : 2012-09-10
copyright            : (C) 2012 by Salvatore Larosa
email                : lrssvtml (at) gmail (dot) com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
Some portions of code were taken from https://code.google.com/p/pydee/
"""

from __future__ import annotations

import importlib
import os
import pyclbr
import re
import sys
import tempfile
from functools import partial
from operator import itemgetter
from pathlib import Path
from typing import TYPE_CHECKING, Optional

from qgis.core import Qgis, QgsApplication, QgsBlockingNetworkRequest, QgsSettings
from qgis.gui import (
    QgsCodeEditorPython,
    QgsCodeEditorWidget,
    QgsGui,
    QgsMessageBar,
    QgsShortcutsManager,
)
from qgis.PyQt.Qsci import QsciScintilla
from qgis.PyQt.QtCore import (
    QByteArray,
    QCoreApplication,
    QDir,
    QEvent,
    QFileInfo,
    QJsonDocument,
    QSize,
    Qt,
    QUrl,
    pyqtSignal,
)
from qgis.PyQt.QtGui import QColor, QKeySequence, QPalette
from qgis.PyQt.QtNetwork import QNetworkRequest
from qgis.PyQt.QtWidgets import (
    QAction,
    QApplication,
    QFileDialog,
    QFrame,
    QGridLayout,
    QLabel,
    QMenu,
    QMessageBox,
    QShortcut,
    QSizePolicy,
    QSpacerItem,
    QTabWidget,
    QToolButton,
    QTreeWidgetItem,
    QWidget,
)
from qgis.utils import OverrideCursor, iface

if TYPE_CHECKING:
    from .console import PythonConsoleWidget


class Editor(QgsCodeEditorPython):
    trigger_find = pyqtSignal()

    def __init__(
        self,
        editor_tab: EditorTab,
        console_widget: PythonConsoleWidget,
        tab_widget: EditorTabWidget,
    ):
        super().__init__(editor_tab)
        self.editor_tab: EditorTab = editor_tab
        self.console_widget: PythonConsoleWidget = console_widget
        self.tab_widget: EditorTabWidget = tab_widget
        self.code_editor_widget: QgsCodeEditorWidget | None = None

        self.setMinimumHeight(120)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)

        # Disable default scintilla shortcuts
        ctrl, shift = self.SCMOD_CTRL << 16, self.SCMOD_SHIFT << 16
        self.SendScintilla(
            QsciScintilla.SCI_CLEARCMDKEY, ord("T") + ctrl
        )  # Switch current line with the next one
        self.SendScintilla(
            QsciScintilla.SCI_CLEARCMDKEY, ord("D") + ctrl
        )  # Duplicate current line / selection
        self.SendScintilla(
            QsciScintilla.SCI_CLEARCMDKEY, ord("L") + ctrl
        )  # Delete current line
        self.SendScintilla(QsciScintilla.SCI_CLEARCMDKEY, ord("L") + ctrl + shift)

        # New QShortcut = ctrl+space/ctrl+alt+space for Autocomplete
        self.newShortcutCS = QShortcut(
            QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_Space), self
        )
        self.newShortcutCS.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.redoScut = QShortcut(
            QKeySequence(Qt.Modifier.CTRL | Qt.Modifier.SHIFT | Qt.Key.Key_Z), self
        )
        self.redoScut.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.redoScut.activated.connect(self.redo)
        self.newShortcutCS.activated.connect(self.autoComplete)
        self.runScriptScut = QShortcut(
            QKeySequence(Qt.Modifier.SHIFT | Qt.Modifier.CTRL | Qt.Key.Key_E), self
        )
        self.runScriptScut.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.runScriptScut.activated.connect(self.runScriptCode)

        self.syntaxCheckScut = QShortcut(
            QKeySequence(Qt.Modifier.CTRL | Qt.Key.Key_4), self
        )
        self.syntaxCheckScut.setContext(Qt.ShortcutContext.WidgetShortcut)
        self.syntaxCheckScut.activated.connect(self.syntaxCheck)
        self.modificationChanged.connect(self.editor_tab.modified)
        self.modificationAttempted.connect(self.fileReadOnly)

    def showApiDocumentation(self, text):
        self.console_widget.shell.showApiDocumentation(text)

    def set_code_editor_widget(self, widget: QgsCodeEditorWidget):
        self.code_editor_widget = widget
        self.code_editor_widget.loadedExternalChanges.connect(
            self.loaded_external_changes
        )

    def settingsEditor(self):
        # Set Python lexer
        self.initializeLexer()

    def contextMenuEvent(self, e):
        menu = QMenu(self)
        menu.addAction(
            QCoreApplication.translate("PythonConsole", "Hide Editor"), self.hideEditor
        )
        menu.addSeparator()

        syntaxCheckAction = QAction(
            QgsApplication.getThemeIcon("console/iconSyntaxErrorConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Check Syntax"),
            menu,
        )
        syntaxCheckAction.triggered.connect(self.syntaxCheck)
        syntaxCheckAction.setShortcut("Ctrl+4")
        menu.addAction(syntaxCheckAction)

        word = self.selectedText() or self.wordAtPoint(e.pos())
        if word:
            context_help_action = QAction(
                QgsApplication.getThemeIcon("mActionHelpContents.svg"),
                QCoreApplication.translate("PythonConsole", "Context Help"),
                menu,
            )
            context_help_action.triggered.connect(
                partial(
                    self.console_widget.shell.showApiDocumentation,
                    word,
                    force_search=True,
                )
            )
            context_help_action.setShortcut(QKeySequence.StandardKey.HelpContents)
            menu.addAction(context_help_action)

        run_selection_action = QAction(menu)
        run_selection_action.setIcon(
            QgsApplication.getThemeIcon("mActionRunSelected.svg"),
        )
        run_selection_action.triggered.connect(self.runSelectedCode)
        QgsGui.shortcutsManager().initializeCommonAction(
            run_selection_action,
            QgsShortcutsManager.CommonAction.CodeRunSelection,
        )
        menu.addAction(run_selection_action)

        start_action = QAction(self)
        start_action.setIcon(QgsApplication.getThemeIcon("mActionStart.svg"))
        start_action.triggered.connect(self.runScriptCode)
        QgsGui.shortcutsManager().initializeCommonAction(
            start_action,
            QgsShortcutsManager.CommonAction.CodeRunScript,
        )
        menu.addAction(start_action)

        menu.addSeparator()
        undoAction = QAction(
            QgsApplication.getThemeIcon("mActionUndo.svg"),
            QCoreApplication.translate("PythonConsole", "Undo"),
            menu,
        )
        undoAction.triggered.connect(self.undo)
        undoAction.setShortcut(QKeySequence.StandardKey.Undo)
        menu.addAction(undoAction)

        redoAction = QAction(
            QgsApplication.getThemeIcon("mActionRedo.svg"),
            QCoreApplication.translate("PythonConsole", "Redo"),
            menu,
        )
        redoAction.triggered.connect(self.redo)
        redoAction.setShortcut("Ctrl+Shift+Z")
        menu.addAction(redoAction)

        menu.addSeparator()
        find_action = QAction(
            QgsApplication.getThemeIcon("console/iconSearchEditorConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Find Text"),
            menu,
        )
        find_action.triggered.connect(self.trigger_find)
        find_action.setShortcut("Ctrl+F")
        menu.addAction(find_action)

        cutAction = QAction(
            QgsApplication.getThemeIcon("mActionEditCut.svg"),
            QCoreApplication.translate("PythonConsole", "Cut"),
            menu,
        )
        cutAction.triggered.connect(self.cut)
        cutAction.setShortcut(QKeySequence.StandardKey.Cut)
        menu.addAction(cutAction)

        copyAction = QAction(
            QgsApplication.getThemeIcon("mActionEditCopy.svg"),
            QCoreApplication.translate("PythonConsole", "Copy"),
            menu,
        )
        copyAction.triggered.connect(self.copy)
        copyAction.setShortcut(QKeySequence.StandardKey.Copy)
        menu.addAction(copyAction)

        pasteAction = QAction(
            QgsApplication.getThemeIcon("mActionEditPaste.svg"),
            QCoreApplication.translate("PythonConsole", "Paste"),
            menu,
        )
        pasteAction.triggered.connect(self.paste)
        pasteAction.setShortcut(QKeySequence.StandardKey.Paste)
        menu.addAction(pasteAction)

        selectAllAction = QAction(
            QCoreApplication.translate("PythonConsole", "Select All"), menu
        )
        selectAllAction.triggered.connect(self.selectAll)
        selectAllAction.setShortcut(QKeySequence.StandardKey.SelectAll)
        menu.addAction(selectAllAction)

        menu.addSeparator()
        toggle_comment_action = QAction(menu)
        toggle_comment_action.setIcon(
            QgsApplication.getThemeIcon(
                "console/iconCommentEditorConsole.svg",
                self.palette().color(QPalette.ColorRole.WindowText),
            )
        )
        toggle_comment_action.triggered.connect(self.toggleComment)
        QgsGui.shortcutsManager().initializeCommonAction(
            toggle_comment_action,
            QgsShortcutsManager.CommonAction.CodeToggleComment,
        )
        menu.addAction(toggle_comment_action)

        reformat_code_action = QAction(menu)
        reformat_code_action.setIcon(
            QgsApplication.getThemeIcon("console/iconFormatCode.svg")
        )
        reformat_code_action.triggered.connect(self.reformatCode)
        QgsGui.shortcutsManager().initializeCommonAction(
            reformat_code_action,
            QgsShortcutsManager.CommonAction.CodeReformat,
        )
        menu.addAction(reformat_code_action)

        menu.addSeparator()
        gist_menu = QMenu(self)
        gist_menu.setTitle(
            QCoreApplication.translate("PythonConsole", "Share on GitHub")
        )
        gist_menu.setIcon(QgsApplication.getThemeIcon("console/iconCodepadConsole.svg"))
        gist_menu.addAction(
            QCoreApplication.translate("PythonConsole", "Secret Gist"),
            partial(self.shareOnGist, False),
        )
        gist_menu.addAction(
            QCoreApplication.translate("PythonConsole", "Public Gist"),
            partial(self.shareOnGist, True),
        )
        menu.addMenu(gist_menu)
        showCodeInspection = menu.addAction(
            QgsApplication.getThemeIcon("console/iconClassBrowserConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Hide/Show Object Inspector"),
            self.objectListEditor,
        )
        menu.addSeparator()
        menu.addAction(
            QgsApplication.getThemeIcon("console/iconSettingsConsole.svg"),
            QCoreApplication.translate("PythonConsole", "Options…"),
            self.console_widget.openSettings,
        )
        syntaxCheckAction.setEnabled(False)
        pasteAction.setEnabled(False)
        cutAction.setEnabled(False)
        run_selection_action.setEnabled(False)
        copyAction.setEnabled(False)
        selectAllAction.setEnabled(False)
        undoAction.setEnabled(False)
        redoAction.setEnabled(False)
        showCodeInspection.setEnabled(False)
        if self.hasSelectedText():
            run_selection_action.setEnabled(True)
            copyAction.setEnabled(True)
            cutAction.setEnabled(True)
        if not self.text() == "":
            selectAllAction.setEnabled(True)
            syntaxCheckAction.setEnabled(True)
        if self.isUndoAvailable():
            undoAction.setEnabled(True)
        if self.isRedoAvailable():
            redoAction.setEnabled(True)
        if QApplication.clipboard().text():
            pasteAction.setEnabled(True)
        if QgsSettings().value("pythonConsole/enableObjectInsp", False, type=bool):
            showCodeInspection.setEnabled(True)
        menu.exec(self.mapToGlobal(e.pos()))

    def objectListEditor(self):
        listObj = self.console_widget.listClassMethod
        if listObj.isVisible():
            listObj.hide()
            self.console_widget.object_inspector_action.setChecked(False)
        else:
            listObj.show()
            self.console_widget.object_inspector_action.setChecked(True)

    def shareOnGist(self, is_public):
        self.code_editor_widget.shareOnGist(is_public)

    def hideEditor(self):
        self.console_widget.splitterObj.hide()
        self.console_widget.show_editor_action.setChecked(False)

    def createTempFile(self):
        # Need to use newline='' to avoid adding extra \r characters on Windows
        with tempfile.NamedTemporaryFile(
            mode="w", encoding="utf-8", newline="", delete=False
        ) as f:
            f.write(self.text())
            name = f.name
        return name

    def runScriptCode(self):
        autoSave = QgsSettings().value("pythonConsole/autoSaveScript", False, type=bool)
        filename = self.code_editor_widget.filePath()
        filename_override = None
        msgEditorBlank = QCoreApplication.translate(
            "PythonConsole", "Empty scripts cannot be run!"
        )
        if not filename:
            if not self.isModified():
                self.showMessage(msgEditorBlank)
                return

        deleteTempFile = False
        if self.syntaxCheck():
            if filename and self.isModified() and autoSave:
                self.save(filename)
            elif not filename or self.isModified():
                # Create a new temp file if the file isn't already saved.
                filename = self.createTempFile()
                filename_override = self.tab_widget.tabText(
                    self.tab_widget.currentIndex()
                )
                if filename_override.startswith("*"):
                    filename_override = filename_override[1:]
                deleteTempFile = True

            self.console_widget.shell.runFile(filename, filename_override)

            if deleteTempFile:
                Path(filename).unlink()

    def runSelectedCode(self):  # spellok
        cmd = self.selectedText()
        self.console_widget.shell.insertFromDropPaste(cmd)
        self.console_widget.shell.entered()
        self.setFocus()

    def getTextFromEditor(self):
        text = self.text()
        textList = text.split("\n")
        return textList

    def goToLine(self, objName, linenr):
        self.SendScintilla(QsciScintilla.SCI_GOTOLINE, linenr - 1)
        self.SendScintilla(
            QsciScintilla.SCI_SETTARGETSTART,
            self.SendScintilla(QsciScintilla.SCI_GETCURRENTPOS),
        )
        self.SendScintilla(QsciScintilla.SCI_SETTARGETEND, len(self.text()))
        pos = self.SendScintilla(
            QsciScintilla.SCI_SEARCHINTARGET, len(objName), objName
        )
        index = pos - self.SendScintilla(QsciScintilla.SCI_GETCURRENTPOS)
        # line, _ = self.getCursorPosition()
        self.setSelection(linenr - 1, index, linenr - 1, index + len(objName))
        self.ensureLineVisible(linenr)
        self.setFocus()

    def syntaxCheck(self):
        self.code_editor_widget.clearWarnings()
        source = self.text().encode("utf-8")
        try:
            compile(source, "", "exec")
        except SyntaxError as detail:
            eline = detail.lineno or 1
            eline -= 1
            ecolumn = detail.offset or 1
            edescr = detail.msg
            self.code_editor_widget.addWarning(eline, edescr)
            self.setCursorPosition(eline, ecolumn - 1)
            self.ensureLineVisible(eline)
            return False

        return True

    def loaded_external_changes(self):
        self.tab_widget.listObject(self.tab_widget.currentWidget())

    def fileReadOnly(self):
        msgText = QCoreApplication.translate(
            "PythonConsole",
            'The file <b>"{0}"</b> is read only, please save to different file first.',
        ).format(self.code_editor_widget.filePath())
        self.showMessage(msgText)

    def save(self, filename: str | None = None):
        if self.isReadOnly():
            return

        if QgsSettings().value("pythonConsole/formatOnSave", False, type=bool):
            self.reformatCode()

        index = self.tab_widget.indexOf(self.editor_tab)
        if not filename and not self.code_editor_widget.filePath():
            saveTr = QCoreApplication.translate(
                "PythonConsole", "Python Console: Save file"
            )
            folder = QgsSettings().value("pythonConsole/lastDirPath", QDir.homePath())
            path, filter = QFileDialog().getSaveFileName(
                self,
                saveTr,
                os.path.join(
                    folder, self.tab_widget.tabText(index).replace("*", "") + ".py"
                ),
                "Script file (*.py)",
            )
            # If the user didn't select a file, abort the save operation
            if not path:
                self.code_editor_widget.setFilePath(None)
                return
            filename = path

        self.code_editor_widget.save(filename)

        msgText = QCoreApplication.translate(
            "PythonConsole", "Script was correctly saved."
        )
        self.showMessage(msgText)

        # Save the new contents
        # Need to use newline='' to avoid adding extra \r characters on Windows
        with open(
            self.code_editor_widget.filePath(), "w", encoding="utf-8", newline=""
        ) as f:
            f.write(self.text())
        self.tab_widget.setTabTitle(
            index, Path(self.code_editor_widget.filePath()).name
        )
        self.tab_widget.setTabToolTip(index, self.code_editor_widget.filePath())
        self.setModified(False)
        self.console_widget.save_file_action.setEnabled(False)
        self.console_widget.updateTabListScript(
            self.code_editor_widget.filePath(), action="append"
        )
        self.tab_widget.listObject(self.editor_tab)
        QgsSettings().setValue(
            "pythonConsole/lastDirPath",
            Path(self.code_editor_widget.filePath()).parent.as_posix(),
        )

    def event(self, e):
        """Used to override the Application shortcuts when the editor has focus"""

        if e.type() == QEvent.Type.ShortcutOverride:
            ctrl = e.modifiers() == Qt.KeyboardModifier.ControlModifier
            ctrl_shift = e.modifiers() == (
                Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.ShiftModifier
            )
            if (
                (ctrl and e.key() == Qt.Key.Key_W)
                or (ctrl_shift and e.key() == Qt.Key.Key_W)
                or (ctrl and e.key() == Qt.Key.Key_S)
                or (ctrl_shift and e.key() == Qt.Key.Key_S)
                or (ctrl and e.key() == Qt.Key.Key_T)
                or (ctrl and e.key() == Qt.Key.Key_Tab)
            ):
                e.accept()
                return True

        return super().event(e)

    def keyPressEvent(self, e):
        ctrl = e.modifiers() == Qt.KeyboardModifier.ControlModifier
        ctrl_shift = e.modifiers() == (
            Qt.KeyboardModifier.ControlModifier | Qt.KeyboardModifier.ShiftModifier
        )

        # Ctrl+W: close current tab
        if ctrl and e.key() == Qt.Key.Key_W:
            self.editor_tab.close()

        # Ctrl+Shift+W: close all tabs
        if ctrl_shift and e.key() == Qt.Key.Key_W:
            self.tab_widget.closeAll()

        # Ctrl+S: save current tab
        if ctrl and e.key() == Qt.Key.Key_S:
            self.save()

        # Ctrl+Shift+S: save current tab as
        if ctrl_shift and e.key() == Qt.Key.Key_S:
            self.tab_widget.saveAs()

        # Ctrl+T: open new tab
        if ctrl and e.key() == Qt.Key.Key_T:
            self.tab_widget.newTabEditor()

        super().keyPressEvent(e)

    def showMessage(
        self, text: str, title: str | None = None, level=Qgis.MessageLevel.Info
    ):
        self.editor_tab.showMessage(text, level, title=title)


class EditorTab(QWidget):
    search_bar_toggled = pyqtSignal(bool)

    def __init__(
        self,
        tab_widget: EditorTabWidget,
        console_widget: PythonConsoleWidget,
        filename: str | None,
        read_only: bool,
    ):
        super().__init__(tab_widget)
        self.tab_widget: EditorTabWidget = tab_widget

        self._editor = Editor(
            editor_tab=self, console_widget=console_widget, tab_widget=tab_widget
        )

        self._editor_code_widget = QgsCodeEditorWidget(self._editor)
        self._editor.set_code_editor_widget(self._editor_code_widget)
        self._editor_code_widget.searchBarToggled.connect(self.search_bar_toggled)

        self._editor.trigger_find.connect(self._editor_code_widget.triggerFind)

        if filename:
            if QFileInfo(filename).exists():
                self._editor_code_widget.loadFile(filename)
                if read_only:
                    self._editor.setReadOnly(True)

        self.tabLayout = QGridLayout(self)
        self.tabLayout.setContentsMargins(0, 0, 0, 0)
        self.tabLayout.addWidget(self._editor_code_widget)

    def set_file_path(self, path: str):
        self._editor_code_widget.setFilePath(path)

    def file_path(self) -> str | None:
        return self._editor_code_widget.filePath()

    def open_in_external_editor(self):
        self._editor_code_widget.openInExternalEditor()

    def modified(self, modified):
        self.tab_widget.tabModified(self, modified)

    def search_bar_visible(self) -> bool:
        """
        Returns True if the tab's search bar is visible
        """
        return self._editor_code_widget.isSearchBarVisible()

    def trigger_find(self):
        """
        Triggers a find operation using the default behavior
        """
        self._editor_code_widget.triggerFind()

    def hide_search_bar(self):
        """
        Hides the search bar
        """
        self._editor_code_widget.hideSearchBar()

    def close(self):
        self.tab_widget._removeTab(self, tab2index=True)

    def __getattr__(self, name):
        """Forward all missing attribute requests to the editor."""
        try:
            return super().__getattr__(name)
        except AttributeError:
            return getattr(self._editor, name)

    def __setattr__(self, name, value):
        """Forward all missing attribute requests to the editor."""
        try:
            return super().__setattr__(name, value)
        except AttributeError:
            return setattr(self._editor, name, value)

    def showMessage(self, text, level=Qgis.MessageLevel.Info, timeout=-1, title=""):
        self._editor_code_widget.messageBar().pushMessage(title, text, level, timeout)


class EditorTabWidget(QTabWidget):
    search_bar_toggled = pyqtSignal(bool)

    def __init__(self, console_widget: PythonConsoleWidget):
        super().__init__(parent=None)
        self.console_widget: PythonConsoleWidget = console_widget

        self.idx = -1
        # Layout for top frame (restore tabs)
        self.layoutTopFrame = QGridLayout(self)
        self.layoutTopFrame.setContentsMargins(0, 0, 0, 0)
        spacerItem = QSpacerItem(
            20, 40, QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Expanding
        )
        self.layoutTopFrame.addItem(spacerItem, 1, 0, 1, 1)
        self.topFrame = QFrame(self)
        self.topFrame.setStyleSheet("background-color: rgb(255, 255, 230);")
        self.topFrame.setFrameShape(QFrame.Shape.StyledPanel)
        self.topFrame.setMinimumHeight(24)
        self.layoutTopFrame2 = QGridLayout(self.topFrame)
        self.layoutTopFrame2.setContentsMargins(0, 0, 0, 0)
        label = QCoreApplication.translate(
            "PythonConsole", "Click on button to restore all tabs from last session."
        )
        self.label = QLabel(label)

        self.restoreTabsButton = QToolButton()
        toolTipRestore = QCoreApplication.translate("PythonConsole", "Restore tabs")
        self.restoreTabsButton.setToolTip(toolTipRestore)
        self.restoreTabsButton.setIcon(
            QgsApplication.getThemeIcon("console/iconRestoreTabsConsole.svg")
        )
        self.restoreTabsButton.setIconSize(QSize(24, 24))
        self.restoreTabsButton.setAutoRaise(True)
        self.restoreTabsButton.setCursor(Qt.CursorShape.PointingHandCursor)
        self.restoreTabsButton.setStyleSheet(
            "QToolButton:hover{border: none } \
                                              QToolButton:pressed{border: none}"
        )

        self.clButton = QToolButton()
        toolTipClose = QCoreApplication.translate("PythonConsole", "Close")
        self.clButton.setToolTip(toolTipClose)
        self.clButton.setIcon(QgsApplication.getThemeIcon("/mIconClose.svg"))
        self.clButton.setIconSize(QSize(18, 18))
        self.clButton.setCursor(Qt.CursorShape.PointingHandCursor)
        self.clButton.setStyleSheet(
            "QToolButton:hover{border: none } \
                                     QToolButton:pressed{border: none}"
        )
        self.clButton.setAutoRaise(True)

        sizePolicy = QSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Fixed)
        self.topFrame.setSizePolicy(sizePolicy)
        self.layoutTopFrame.addWidget(self.topFrame, 0, 0, 1, 1)
        self.layoutTopFrame2.addWidget(self.label, 0, 1, 1, 1)
        self.layoutTopFrame2.addWidget(self.restoreTabsButton, 0, 0, 1, 1)
        self.layoutTopFrame2.addWidget(self.clButton, 0, 2, 1, 1)

        self.topFrame.hide()
        self.restoreTabsButton.clicked.connect(self.restoreTabs)
        self.clButton.clicked.connect(self.closeRestore)

        # Fixes #7653
        if sys.platform != "darwin":
            self.setDocumentMode(True)

        self.setMovable(True)
        self.setTabsClosable(True)
        self.setTabPosition(QTabWidget.TabPosition.North)

        # Menu button list tabs
        self.fileTabMenu = QMenu()
        self.fileTabMenu.aboutToShow.connect(self.showFileTabMenu)
        self.fileTabMenu.triggered.connect(self.showFileTabMenuTriggered)
        self.fileTabButton = QToolButton()
        txtToolTipMenuFile = QCoreApplication.translate(
            "PythonConsole", "List all tabs"
        )
        self.fileTabButton.setToolTip(txtToolTipMenuFile)
        self.fileTabButton.setIcon(
            QgsApplication.getThemeIcon("console/iconFileTabsMenuConsole.svg")
        )
        self.fileTabButton.setIconSize(QSize(24, 24))
        self.fileTabButton.setAutoRaise(True)
        self.fileTabButton.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.fileTabButton.setMenu(self.fileTabMenu)
        self.setCornerWidget(self.fileTabButton, Qt.Corner.TopRightCorner)
        self.tabCloseRequested.connect(self._removeTab)
        self.currentChanged.connect(self._currentWidgetChanged)

        # New Editor button
        self.newTabButton = QToolButton()
        txtToolTipNewTab = QCoreApplication.translate("PythonConsole", "New Editor")
        self.newTabButton.setToolTip(f"<b>{txtToolTipNewTab}</b> (Ctrl+T)")
        self.newTabButton.setAutoRaise(True)
        self.newTabButton.setIcon(
            QgsApplication.getThemeIcon("console/iconNewTabEditorConsole.svg")
        )
        self.newTabButton.setIconSize(QSize(24, 24))
        self.setCornerWidget(self.newTabButton, Qt.Corner.TopLeftCorner)
        self.newTabButton.clicked.connect(self.newTabEditor)

    def _currentWidgetChanged(self, tab):
        if QgsSettings().value("pythonConsole/enableObjectInsp", False, type=bool):
            self.listObject(tab)
        self.changeLastDirPath(tab)
        self.enableSaveIfModified(tab)

        if self.currentWidget():
            self.search_bar_toggled.emit(self.currentWidget().search_bar_visible())

    def toggle_search_bar(self, visible: bool):
        """
        Toggles whether the search bar should be visible
        """
        if visible and not self.currentWidget().search_bar_visible():
            self.currentWidget().trigger_find()
        elif not visible and self.currentWidget().search_bar_visible():
            self.currentWidget().hide_search_bar()

    def contextMenuEvent(self, e):
        tabBar = self.tabBar()
        self.idx = tabBar.tabAt(e.pos())
        if self.widget(self.idx):
            cW = self.widget(self.idx)
            menu = QMenu(self)
            menu.addSeparator()
            menu.addAction(
                QCoreApplication.translate("PythonConsole", "New Editor"),
                self.newTabEditor,
            )
            menu.addSeparator()
            closeTabAction = menu.addAction(
                QCoreApplication.translate("PythonConsole", "Close Tab"), cW.close
            )
            closeAllTabAction = menu.addAction(
                QCoreApplication.translate("PythonConsole", "Close All"), self.closeAll
            )
            closeOthersTabAction = menu.addAction(
                QCoreApplication.translate("PythonConsole", "Close Others"),
                self.closeOthers,
            )
            menu.addSeparator()
            saveAction = menu.addAction(
                QCoreApplication.translate("PythonConsole", "Save"), cW.save
            )
            menu.addAction(
                QCoreApplication.translate("PythonConsole", "Save As"), self.saveAs
            )
            closeTabAction.setEnabled(False)
            closeAllTabAction.setEnabled(False)
            closeOthersTabAction.setEnabled(False)
            saveAction.setEnabled(False)
            if self.count() > 1:
                closeTabAction.setEnabled(True)
                closeAllTabAction.setEnabled(True)
                closeOthersTabAction.setEnabled(True)
            if self.widget(self.idx).isModified():
                saveAction.setEnabled(True)
            menu.exec(self.mapToGlobal(e.pos()))

    def closeOthers(self):
        idx = self.idx
        countTab = self.count()
        for i in list(range(countTab - 1, idx, -1)) + list(range(idx - 1, -1, -1)):
            self._removeTab(i)

    def closeAll(self):
        countTab = self.count()
        for i in range(countTab - 1, 0, -1):
            self._removeTab(i)

        self.newTabEditor(
            tabName=QCoreApplication.translate("PythonConsole", "Untitled-0")
        )
        self._removeTab(0)

    def saveAs(self):
        self.console_widget.saveAsScriptFile(self.idx)

    def enableSaveIfModified(self, tab):
        tabWidget = self.widget(tab)
        if tabWidget:
            self.console_widget.save_file_action.setEnabled(tabWidget.isModified())

    def enableToolBarEditor(self, enable):
        if self.topFrame.isVisible():
            enable = False
        self.console_widget.toolBarEditor.setEnabled(enable)

    def newTabEditor(self, tabName=None, filename: str | None = None):
        read_only = False
        if filename:
            read_only = not QFileInfo(filename).isWritable()
            try:
                fn = open(filename, "rb")
                fn.read()
                fn.close()
            except OSError as error:
                IOErrorTr = QCoreApplication.translate(
                    "PythonConsole", "The file {0} could not be opened. Error: {1}\n"
                ).format(filename, error.strerror)
                print("## Error: ")
                sys.stderr.write(IOErrorTr)
                return

        nr = self.count()
        if not tabName:
            tabName = QCoreApplication.translate(
                "PythonConsole", "Untitled-{0}"
            ).format(nr)
        tab = EditorTab(
            tab_widget=self,
            console_widget=self.console_widget,
            filename=filename,
            read_only=read_only,
        )
        self.iconTab = QgsApplication.getThemeIcon("console/iconTabEditorConsole.svg")
        self.addTab(tab, self.iconTab, tabName + " (ro)" if read_only else tabName)
        self.setCurrentWidget(tab)
        if filename:
            self.setTabToolTip(self.currentIndex(), filename)
        else:
            self.setTabToolTip(self.currentIndex(), tabName)

        tab.search_bar_toggled.connect(self._tab_search_bar_toggled)

    def _tab_search_bar_toggled(self, visible: bool):
        if self.sender() != self.currentWidget():
            return

        self.search_bar_toggled.emit(visible)

    def tabModified(self, tab, modified):
        index = self.indexOf(tab)
        s = self.tabText(index)
        self.setTabTitle(index, f"*{s}" if modified else re.sub(r"^(\*)", "", s))
        self.console_widget.save_file_action.setEnabled(modified)

    def setTabTitle(self, tab, title):
        self.setTabText(tab, title)

    def _removeTab(self, tab, tab2index=False):
        if tab2index:
            tab = self.indexOf(tab)
        editorTab = self.widget(tab)
        if editorTab.isModified():
            txtSaveOnRemove = QCoreApplication.translate(
                "PythonConsole", "Python Console: Save File"
            )
            txtMsgSaveOnRemove = QCoreApplication.translate(
                "PythonConsole",
                "The file <b>'{0}'</b> has been modified, save changes?",
            ).format(self.tabText(tab))
            res = QMessageBox.question(
                self,
                txtSaveOnRemove,
                txtMsgSaveOnRemove,
                QMessageBox.StandardButton.Save
                | QMessageBox.StandardButton.Discard
                | QMessageBox.StandardButton.Cancel,
            )
            if res == QMessageBox.StandardButton.Cancel:
                return
            if res == QMessageBox.StandardButton.Save:
                editorTab.save()
            if editorTab.code_editor_widget.filePath():
                self.console_widget.updateTabListScript(
                    editorTab.code_editor_widget.filePath(), action="remove"
                )
            self.removeTab(tab)
            if self.count() < 1:
                self.newTabEditor()
        else:
            if editorTab.code_editor_widget.filePath():
                self.console_widget.updateTabListScript(
                    editorTab.code_editor_widget.filePath(), action="remove"
                )
            if self.count() <= 1:
                self.removeTab(tab)
                self.newTabEditor()
            else:
                self.removeTab(tab)

        editorTab.deleteLater()
        self.currentWidget()._editor.setFocus(Qt.FocusReason.TabFocusReason)

    def restoreTabsOrAddNew(self):
        """
        Restore tabs if they are found in the settings. If none are found it will add a new empty tab.
        """
        # Restore scripts from the previous session
        tabScripts = QgsSettings().value("pythonConsole/tabScripts", [])
        self.restoreTabList = tabScripts

        if self.restoreTabList:
            self.restoreTabs()
        else:
            self.newTabEditor(filename=None)

    def restoreTabs(self):
        for script in self.restoreTabList:
            pathFile = script
            if QFileInfo(pathFile).exists():
                tabName = pathFile.split("/")[-1]
                self.newTabEditor(tabName, pathFile)
            else:
                errOnRestore = QCoreApplication.translate(
                    "PythonConsole", "Unable to restore the file: \n{0}\n"
                ).format(pathFile)
                print("## Error: ")
                s = errOnRestore
                sys.stderr.write(s)
                self.console_widget.updateTabListScript(pathFile, action="remove")
        if self.count() < 1:
            self.newTabEditor(filename=None)
        self.topFrame.close()
        self.enableToolBarEditor(True)
        self.currentWidget()._editor.setFocus(Qt.FocusReason.TabFocusReason)

    def closeRestore(self):
        self.console_widget.updateTabListScript(None)
        self.topFrame.close()
        self.newTabEditor(filename=None)
        self.enableToolBarEditor(True)

    def showFileTabMenu(self):
        self.fileTabMenu.clear()
        for index in range(self.count()):
            action = self.fileTabMenu.addAction(
                self.tabIcon(index), self.tabText(index)
            )
            action.setData(index)

    def showFileTabMenuTriggered(self, action):
        index = action.data()
        if index is not None:
            self.setCurrentIndex(index)

    def listObject(self, tab):
        self.console_widget.listClassMethod.clear()
        if isinstance(tab, EditorTab):
            tabWidget = self.widget(self.indexOf(tab))
        else:
            tabWidget = self.widget(tab)
        if tabWidget:
            if tabWidget.file_path():
                pathFile, file = os.path.split(tabWidget.file_path())
                module, ext = os.path.splitext(file)
                found = False
                if pathFile not in sys.path:
                    sys.path.append(pathFile)
                    found = True
                try:
                    importlib.reload(pyclbr)  # NOQA
                    dictObject = {}
                    readModule = pyclbr.readmodule(module)
                    readModuleFunction = pyclbr.readmodule_ex(module)
                    for name, class_data in sorted(
                        list(readModule.items()), key=lambda x: x[1].lineno
                    ):
                        if os.path.normpath(class_data.file) == os.path.normpath(
                            tabWidget.file_path()
                        ):
                            superClassName = []
                            for superClass in class_data.super:
                                if superClass == "object":
                                    continue
                                if isinstance(superClass, str):
                                    superClassName.append(superClass)
                                else:
                                    superClassName.append(superClass.name)
                            classItem = QTreeWidgetItem()
                            if superClassName:
                                super = ", ".join([i for i in superClassName])
                                classItem.setText(0, name + " [" + super + "]")
                                classItem.setToolTip(0, name + " [" + super + "]")
                            else:
                                classItem.setText(0, name)
                                classItem.setToolTip(0, name)
                            if sys.platform.startswith("win"):
                                classItem.setSizeHint(0, QSize(18, 18))
                            classItem.setText(1, str(class_data.lineno))
                            iconClass = QgsApplication.getThemeIcon(
                                "console/iconClassTreeWidgetConsole.svg"
                            )
                            classItem.setIcon(0, iconClass)
                            dictObject[name] = class_data.lineno
                            for meth, lineno in sorted(
                                list(class_data.methods.items()), key=itemgetter(1)
                            ):
                                methodItem = QTreeWidgetItem()
                                methodItem.setText(0, meth + " ")
                                methodItem.setText(1, str(lineno))
                                methodItem.setToolTip(0, meth)
                                iconMeth = QgsApplication.getThemeIcon(
                                    "console/iconMethodTreeWidgetConsole.svg"
                                )
                                methodItem.setIcon(0, iconMeth)
                                if sys.platform.startswith("win"):
                                    methodItem.setSizeHint(0, QSize(18, 18))
                                classItem.addChild(methodItem)
                                dictObject[meth] = lineno
                            self.console_widget.listClassMethod.addTopLevelItem(
                                classItem
                            )
                    for func_name, data in sorted(
                        list(readModuleFunction.items()), key=lambda x: x[1].lineno
                    ):
                        if isinstance(data, pyclbr.Function) and os.path.normpath(
                            data.file
                        ) == os.path.normpath(tabWidget.file_path()):
                            funcItem = QTreeWidgetItem()
                            funcItem.setText(0, func_name + " ")
                            funcItem.setText(1, str(data.lineno))
                            funcItem.setToolTip(0, func_name)
                            iconFunc = QgsApplication.getThemeIcon(
                                "console/iconFunctionTreeWidgetConsole.svg"
                            )
                            funcItem.setIcon(0, iconFunc)
                            if sys.platform.startswith("win"):
                                funcItem.setSizeHint(0, QSize(18, 18))
                            dictObject[func_name] = data.lineno
                            self.console_widget.listClassMethod.addTopLevelItem(
                                funcItem
                            )
                    if found:
                        sys.path.remove(pathFile)
                except:
                    msgItem = QTreeWidgetItem()
                    msgItem.setText(
                        0, QCoreApplication.translate("PythonConsole", "Check Syntax")
                    )
                    msgItem.setText(1, "syntaxError")
                    iconWarning = QgsApplication.getThemeIcon(
                        "console/iconSyntaxErrorConsole.svg"
                    )
                    msgItem.setIcon(0, iconWarning)
                    self.console_widget.listClassMethod.addTopLevelItem(msgItem)

    def refreshSettingsEditor(self):
        objInspectorEnabled = QgsSettings().value(
            "pythonConsole/enableObjectInsp", False, type=bool
        )
        if self.console_widget.listClassMethod.isVisible():
            self.console_widget.object_inspector_action.setChecked(objInspectorEnabled)
        self.console_widget.object_inspector_action.setEnabled(objInspectorEnabled)
        if objInspectorEnabled:
            cW = self.currentWidget()
            if cW and not self.console_widget.listClassMethod.isVisible():
                with OverrideCursor(Qt.CursorShape.WaitCursor):
                    self.listObject(cW)

    def changeLastDirPath(self, tab):
        tabWidget = self.widget(tab)
        if tabWidget and tabWidget.file_path():
            QgsSettings().setValue(
                "pythonConsole/lastDirPath",
                Path(tabWidget.file_path()).parent.as_posix(),
            )

    def showMessage(self, text, level=Qgis.MessageLevel.Info, timeout=-1, title=""):
        currWidget = self.currentWidget()
        currWidget.showMessage(text, level, timeout, title)



"""
/***************************************************************************
Python Console for QGIS
                             -------------------
begin                : 2012-09-10
copyright            : (C) 2012 by Salvatore Larosa
email                : lrssvtml (at) gmail (dot) com
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
Some portions of code were taken from https://code.google.com/p/pydee/
"""

import os
import re
import subprocess
import sys
from functools import partial

from qgis.core import Qgis, QgsApplication, QgsFileUtils, QgsSettings
from qgis.gui import (
    QgsApplicationExitBlockerInterface,
    QgsCodeEditorDockWidget,
    QgsDockWidget,
    QgsFilterLineEdit,
    QgsGui,
    QgsHelp,
    QgsShortcutsManager,
)
from qgis.PyQt.QtCore import (
    QByteArray,
    QCoreApplication,
    QDir,
    QFileInfo,
    QSize,
    Qt,
    QTimer,
    QUrl,
)
from qgis.PyQt.QtGui import QColor, QDesktopServices, QKeySequence, QPalette
from qgis.PyQt.QtWidgets import (
    QAction,
    QApplication,
    QCheckBox,
    QFileDialog,
    QGridLayout,
    QMenu,
    QMessageBox,
    QShortcut,
    QSizePolicy,
    QSplitter,
    QToolBar,
    QToolButton,
    QTreeWidget,
    QVBoxLayout,
    QWidget,
)
from qgis.utils import iface

from .console_editor import EditorTabWidget
from .console_output import ShellOutputScintilla
from .console_sci import ShellScintilla
from .console_settings import ConsoleOptionsFactory

_console = None
_options_factory = ConsoleOptionsFactory()


def show_console():
    """called from QGIS to open the console"""
    global _console
    if _console is None:
        parent = iface.mainWindow() if iface else None
        _console = PythonConsole(parent)
        if iface:
            _console.visibilityChanged.connect(
                iface.actionShowPythonDialog().setChecked
            )

        _console.show()  # force show even if it was restored as hidden
        # set focus to the console so the user can start typing
        # defer the set focus event so it works also whether the console not visible yet
        QTimer.singleShot(0, _console.activate)
    else:
        _console.setUserVisible(not _console.isUserVisible())
        # set focus to the console so the user can start typing
        if _console.isUserVisible():
            _console.activate()

    return _console


_console_output = None

# hook for python console so all output will be redirected
# and then shown in console


def console_displayhook(obj):
    global _console_output
    _console_output = obj


def init_console():
    """
    Called from QGIS to initialize the console related options and shortcuts,
    before the dock is shown
    """
    global _options_factory
    _options_factory.setTitle(QCoreApplication.translate("PythonConsole", "Python"))
    iface.registerOptionsWidgetFactory(_options_factory)


class ConsoleExitBlocker(QgsApplicationExitBlockerInterface):
    def __init__(self, console):
        super().__init__()
        self.console = console

    def allowExit(self):
        return self.console.allowExit()


class PythonConsole(QgsCodeEditorDockWidget):
    def __init__(self, parent=None):
        super().__init__("PythonConsoleWindow", True)
        self.setDockObjectName("PythonConsole")
        self.setTitle(QCoreApplication.translate("PythonConsole", "Python Console"))

        self.console = PythonConsoleWidget(self)
        QgsGui.instance().optionsChanged.connect(self.console.updateSettings)
        vl = QVBoxLayout()
        vl.setContentsMargins(0, 0, 0, 0)
        vl.addWidget(self.console)
        self.setLayout(vl)
        self.setFocusProxy(self.console)

        # closeEvent is not always called for this widget -- so we also trigger a settings
        # save on application exit
        QgsApplication.instance().aboutToQuit.connect(self.on_app_exit)

    def on_app_exit(self):
        self.console.saveSettingsConsole()
        self.deleteLater()

    def activate(self):
        self.activateWindow()
        self.raise_()
        self.setFocus()

    def closeEvent(self, event):
        self.console.saveSettingsConsole()
        self.hide()
        event.ignore()


class PythonConsoleWidget(QWidget):
    def __init__(self, parent=None):
        QWidget.__init__(self, parent)
        self.setWindowTitle(
            QCoreApplication.translate("PythonConsole", "Python Console")
        )

        self.shell = ShellScintilla(console_widget=self)
        self.setFocusProxy(self.shell)
        self.shell_output = ShellOutputScintilla(
            console_widget=self, shell_editor=self.shell
        )
        self.tabEditorWidget = EditorTabWidget(console_widget=self)

        # ------------ UI -------------------------------

        self.splitterEditor = QSplitter(self)
        self.splitterEditor.setOrientation(Qt.Orientation.Horizontal)
        self.splitterEditor.setHandleWidth(6)
        self.splitterEditor.setChildrenCollapsible(True)

        self.shellOutWidget = QWidget(self)
        self.shellOutWidget.setLayout(QVBoxLayout())
        self.shellOutWidget.layout().setContentsMargins(0, 0, 0, 0)
        self.shellOutWidget.layout().addWidget(self.shell_output)

        self.splitter = QSplitter(self.splitterEditor)
        self.splitter.setOrientation(Qt.Orientation.Vertical)
        self.splitter.setHandleWidth(3)
        self.splitter.setChildrenCollapsible(False)
        self.splitter.addWidget(self.shellOutWidget)
        self.splitter.addWidget(self.shell)

        self.splitterObj = QSplitter(self.splitterEditor)
        self.splitterObj.setHandleWidth(3)
        self.splitterObj.setOrientation(Qt.Orientation.Horizontal)

        self.widgetEditor = QWidget(self.splitterObj)

        self.listClassMethod = QTreeWidget(self.splitterObj)
        self.listClassMethod.setColumnCount(2)
        objInspLabel = QCoreApplication.translate("PythonConsole", "Object Inspector")
        self.listClassMethod.setHeaderLabels([objInspLabel, ""])
        self.listClassMethod.setColumnHidden(1, True)
        self.listClassMethod.setAlternatingRowColors(True)

        # Hide side editor on start up
        self.splitterObj.hide()
        self.listClassMethod.hide()

        icon_size = iface.iconSize(dockedToolbar=True) if iface else QSize(16, 16)

        sizes = self.splitter.sizes()
        self.splitter.setSizes(sizes)

        # ----------------Restore Settings------------------------------------

        self.restoreSettingsConsole()

        # ------------------Toolbar Editor-------------------------------------

        # Action for Open File
        openFileBt = QCoreApplication.translate("PythonConsole", "Open Script…")
        self.open_file_action = QAction(self)
        self.open_file_action.setIcon(
            QgsApplication.getThemeIcon("mActionScriptOpen.svg")
        )
        self.open_file_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.open_file_action.setIconVisibleInMenu(True)
        self.open_file_action.setToolTip(f"<b>{openFileBt}</b> (Ctrl+O)")
        self.open_file_action.setText(openFileBt)

        openExtEditorBt = QCoreApplication.translate(
            "PythonConsole", "Open in External Editor"
        )
        self.open_in_editor_action = QAction(self)
        self.open_in_editor_action.setIcon(
            QgsApplication.getThemeIcon("console/iconShowEditorConsole.svg")
        )
        self.open_in_editor_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.open_in_editor_action.setIconVisibleInMenu(True)
        self.open_in_editor_action.setToolTip(openExtEditorBt)
        self.open_in_editor_action.setText(openExtEditorBt)

        # Action for Save File
        saveFileBt = QCoreApplication.translate("PythonConsole", "Save")
        self.save_file_action = QAction(self)
        self.save_file_action.setEnabled(False)
        self.save_file_action.setIcon(
            QgsApplication.getThemeIcon("mActionFileSave.svg")
        )
        self.save_file_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.save_file_action.setIconVisibleInMenu(True)
        self.save_file_action.setToolTip(f"<b>{saveFileBt}</b> (Ctrl+S)")
        self.save_file_action.setText(saveFileBt)

        # Action for Save File As
        saveAsFileBt = QCoreApplication.translate("PythonConsole", "Save As…")
        self.save_as_file_action = QAction(self)
        self.save_as_file_action.setIcon(
            QgsApplication.getThemeIcon("mActionFileSaveAs.svg")
        )
        self.save_as_file_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.save_as_file_action.setIconVisibleInMenu(True)
        self.save_as_file_action.setToolTip(f"<b>{saveAsFileBt}</b> (Ctrl+Shift+S)")
        self.save_as_file_action.setText(saveAsFileBt)

        # Action Cut
        cutEditorBt = QCoreApplication.translate("PythonConsole", "Cut")
        self.cut_action = QAction(self)
        self.cut_action.setIcon(QgsApplication.getThemeIcon("mActionEditCut.svg"))
        self.cut_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.cut_action.setIconVisibleInMenu(True)
        self.cut_action.setToolTip(f"<b>{cutEditorBt}</b> (Ctrl+X)")
        self.cut_action.setText(cutEditorBt)

        # Action Copy
        copyEditorBt = QCoreApplication.translate("PythonConsole", "Copy")
        self.copy_action = QAction(self)
        self.copy_action.setIcon(QgsApplication.getThemeIcon("mActionEditCopy.svg"))
        self.copy_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.copy_action.setIconVisibleInMenu(True)
        self.copy_action.setToolTip(f"<b>{copyEditorBt}</b> (Ctrl+C)")
        self.copy_action.setText(copyEditorBt)

        # Action Paste
        pasteEditorBt = QCoreApplication.translate("PythonConsole", "Paste")
        self.paste_action = QAction(self)
        self.paste_action.setIcon(QgsApplication.getThemeIcon("mActionEditPaste.svg"))
        self.paste_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.paste_action.setIconVisibleInMenu(True)
        self.paste_action.setToolTip(f"<b>{pasteEditorBt}</b> (Ctrl+V)")
        self.paste_action.setText(pasteEditorBt)

        # Action Run Script (subprocess)
        self.run_script_action = QAction(self)
        self.run_script_action.setIcon(QgsApplication.getThemeIcon("mActionStart.svg"))
        self.run_script_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.run_script_action.setIconVisibleInMenu(True)
        QgsGui.shortcutsManager().initializeCommonAction(
            self.run_script_action, QgsShortcutsManager.CommonAction.CodeRunScript
        )

        # Action Run Selected
        self.run_selection_action = QAction(self)
        self.run_selection_action.setIcon(
            QgsApplication.getThemeIcon("mActionRunSelected.svg")
        )
        self.run_selection_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.run_selection_action.setIconVisibleInMenu(True)
        QgsGui.shortcutsManager().initializeCommonAction(
            self.run_selection_action, QgsShortcutsManager.CommonAction.CodeRunSelection
        )

        # Action Toggle comment
        self.toggle_comment_action = QAction(self)
        self.toggle_comment_action.setIcon(
            QgsApplication.getThemeIcon(
                "console/iconCommentEditorConsole.svg",
                self.palette().color(QPalette.ColorRole.WindowText),
            ),
        )
        self.toggle_comment_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.toggle_comment_action.setIconVisibleInMenu(True)
        QgsGui.shortcutsManager().initializeCommonAction(
            self.toggle_comment_action,
            QgsShortcutsManager.CommonAction.CodeToggleComment,
        )

        # Action Format code
        self.reformat_code_action = QAction(self)
        self.reformat_code_action.setIcon(
            QgsApplication.getThemeIcon("console/iconFormatCode.svg")
        )
        self.reformat_code_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.reformat_code_action.setIconVisibleInMenu(True)
        QgsGui.shortcutsManager().initializeCommonAction(
            self.reformat_code_action, QgsShortcutsManager.CommonAction.CodeReformat
        )

        # Action for Object browser
        objList = QCoreApplication.translate("PythonConsole", "Object Inspector…")
        self.object_inspector_action = QAction(self)
        self.object_inspector_action.setCheckable(True)
        self.object_inspector_action.setEnabled(
            QgsSettings().value("pythonConsole/enableObjectInsp", False, type=bool)
        )
        self.object_inspector_action.setIcon(
            QgsApplication.getThemeIcon("console/iconClassBrowserConsole.svg")
        )
        self.object_inspector_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.object_inspector_action.setIconVisibleInMenu(True)
        self.object_inspector_action.setToolTip(objList)
        self.object_inspector_action.setText(objList)

        # Action for Find text
        findText = QCoreApplication.translate("PythonConsole", "Find Text")
        self.find_text_action = QAction(self)
        self.find_text_action.setCheckable(True)
        self.find_text_action.setEnabled(True)
        self.find_text_action.setIcon(
            QgsApplication.getThemeIcon("console/iconSearchEditorConsole.svg")
        )
        self.find_text_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.find_text_action.setIconVisibleInMenu(True)
        self.find_text_action.setToolTip(f"<b>{findText}</b> (Ctrl+F)")
        self.find_text_action.setText(findText)

        self.tabEditorWidget.search_bar_toggled.connect(
            self.find_text_action.setChecked
        )
        self.find_text_action.toggled.connect(self.tabEditorWidget.toggle_search_bar)

        # ----------------Toolbar Console-------------------------------------

        # Action Show Editor
        showEditor = QCoreApplication.translate("PythonConsole", "Show Editor")
        self.show_editor_action = QAction(self)
        self.show_editor_action.setCheckable(True)
        self.show_editor_action.setIcon(
            QgsApplication.getThemeIcon("console/iconShowEditorConsole.svg")
        )
        self.show_editor_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.show_editor_action.setIconVisibleInMenu(True)
        self.show_editor_action.setToolTip(showEditor)
        self.show_editor_action.setText(showEditor)

        # Action for Clear button
        clearBt = QCoreApplication.translate("PythonConsole", "Clear Console")
        self.clear_action = QAction(self)
        self.clear_action.setIcon(
            QgsApplication.getThemeIcon("console/iconClearConsole.svg")
        )
        self.clear_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.clear_action.setIconVisibleInMenu(True)
        self.clear_action.setToolTip(clearBt)
        self.clear_action.setText(clearBt)

        # Action for settings
        optionsBt = QCoreApplication.translate("PythonConsole", "Options…")
        self.options_action = QAction(self)
        self.options_action.setIcon(
            QgsApplication.getThemeIcon("console/iconSettingsConsole.svg")
        )
        self.options_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.options_action.setIconVisibleInMenu(True)
        self.options_action.setToolTip(optionsBt)
        self.options_action.setText(optionsBt)

        # Action for Run script
        runBt = QCoreApplication.translate("PythonConsole", "Run Command")
        self.run_action = QAction(self)
        self.run_action.setIcon(QgsApplication.getThemeIcon("mActionStart.svg"))
        self.run_action.setMenuRole(QAction.MenuRole.PreferencesRole)
        self.run_action.setIconVisibleInMenu(True)
        self.run_action.setToolTip(runBt)
        self.run_action.setText(runBt)

        # Help button
        self.help_console_action = QAction(self)
        self.help_console_action.setText(
            QCoreApplication.translate("PythonConsole", "Python Console Help")
        )
        self.help_api_action = QAction(self)
        self.help_api_action.setText(
            QCoreApplication.translate("PythonConsole", "PyQGIS API Documentation")
        )
        self.help_cookbook_action = QAction(self)
        self.help_cookbook_action.setText(
            QCoreApplication.translate("PythonConsole", "PyQGIS Cookbook")
        )

        self.helpMenu = QMenu(self)
        self.helpMenu.addAction(self.help_console_action)
        self.helpMenu.addAction(self.help_api_action)
        self.helpMenu.addAction(self.help_cookbook_action)

        helpBt = QCoreApplication.translate("PythonConsole", "Help…")
        self.helpButton = QToolButton(self)
        self.helpButton.setPopupMode(QToolButton.ToolButtonPopupMode.InstantPopup)
        self.helpButton.setEnabled(True)
        self.helpButton.setIcon(
            QgsApplication.getThemeIcon("console/iconHelpConsole.svg")
        )
        self.helpButton.setToolTip(helpBt)
        self.helpButton.setMenu(self.helpMenu)

        self.toolBar = QToolBar()
        self.toolBar.setEnabled(True)
        self.toolBar.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.toolBar.setContextMenuPolicy(Qt.ContextMenuPolicy.DefaultContextMenu)
        self.toolBar.setLayoutDirection(Qt.LayoutDirection.LeftToRight)
        self.toolBar.setIconSize(icon_size)
        self.toolBar.setMovable(False)
        self.toolBar.setFloatable(False)
        self.toolBar.addAction(self.clear_action)
        self.toolBar.addAction(self.run_action)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.show_editor_action)
        self.toolBar.addSeparator()
        self.toolBar.addAction(self.options_action)
        self.toolBar.addWidget(self.helpButton)
        self.toolBar.addSeparator()
        self.toolBar.addWidget(parent.dockToggleButton())

        self.toolBarEditor = QToolBar()
        self.toolBarEditor.setEnabled(False)
        self.toolBarEditor.setFocusPolicy(Qt.FocusPolicy.NoFocus)
        self.toolBarEditor.setContextMenuPolicy(Qt.ContextMenuPolicy.DefaultContextMenu)
        self.toolBarEditor.setLayoutDirection(Qt.LayoutDirection.LeftToRight)
        self.toolBarEditor.setIconSize(icon_size)
        self.toolBarEditor.setMovable(False)
        self.toolBarEditor.setFloatable(False)
        self.toolBarEditor.addAction(self.open_file_action)
        self.toolBarEditor.addAction(self.open_in_editor_action)
        self.toolBarEditor.addSeparator()
        self.toolBarEditor.addAction(self.save_file_action)
        self.toolBarEditor.addAction(self.save_as_file_action)
        self.toolBarEditor.addSeparator()
        self.toolBarEditor.addAction(self.run_script_action)
        self.toolBarEditor.addAction(self.run_selection_action)
        self.toolBarEditor.addSeparator()
        self.toolBarEditor.addAction(self.cut_action)
        self.toolBarEditor.addAction(self.copy_action)
        self.toolBarEditor.addAction(self.paste_action)
        self.toolBarEditor.addSeparator()
        self.toolBarEditor.addAction(self.find_text_action)
        self.toolBarEditor.addSeparator()
        self.toolBarEditor.addAction(self.toggle_comment_action)
        self.toolBarEditor.addAction(self.reformat_code_action)
        self.toolBarEditor.addSeparator()
        self.toolBarEditor.addAction(self.object_inspector_action)

        self.widgetButton = QWidget()
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.widgetButton.sizePolicy().hasHeightForWidth())
        self.widgetButton.setSizePolicy(sizePolicy)

        self.widgetButtonEditor = QWidget(self.widgetEditor)
        sizePolicy = QSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.Preferred)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(
            self.widgetButtonEditor.sizePolicy().hasHeightForWidth()
        )
        self.widgetButtonEditor.setSizePolicy(sizePolicy)

        sizePolicy = QSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.shell_output.sizePolicy().hasHeightForWidth())
        self.shell_output.setSizePolicy(sizePolicy)

        self.shell_output.setVerticalScrollBarPolicy(
            Qt.ScrollBarPolicy.ScrollBarAsNeeded
        )
        self.shell.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAsNeeded)

        # ------------ Layout -------------------------------

        self.mainLayout = QGridLayout(self)
        self.mainLayout.setMargin(0)
        self.mainLayout.setSpacing(0)
        self.mainLayout.addWidget(self.widgetButton, 0, 0, 1, 1)
        self.mainLayout.addWidget(self.splitterEditor, 0, 1, 1, 1)

        self.shellOutWidget.layout().insertWidget(0, self.toolBar)

        self.layoutEditor = QGridLayout(self.widgetEditor)
        self.layoutEditor.setMargin(0)
        self.layoutEditor.setSpacing(0)
        self.layoutEditor.addWidget(self.toolBarEditor, 0, 1, 1, 1)
        self.layoutEditor.addWidget(self.widgetButtonEditor, 1, 0, 2, 1)
        self.layoutEditor.addWidget(self.tabEditorWidget, 1, 1, 1, 1)

        # ------------ Add first Tab in Editor -------------------------------

        # self.tabEditorWidget.newTabEditor(tabName='first', filename=None)

        # ------------ Signal -------------------------------

        self.object_inspector_action.toggled.connect(self.toggleObjectListWidget)
        self.toggle_comment_action.triggered.connect(self.toggleComment)
        self.reformat_code_action.triggered.connect(self.reformatCode)
        self.run_script_action.triggered.connect(self.runScriptEditor)
        self.run_selection_action.triggered.connect(self.runSelectedEditor)
        self.cut_action.triggered.connect(self.cutEditor)
        self.copy_action.triggered.connect(self.copyEditor)
        self.paste_action.triggered.connect(self.pasteEditor)
        self.show_editor_action.toggled.connect(self.toggleEditor)
        self.clear_action.triggered.connect(self.shell_output.clearConsole)
        self.options_action.triggered.connect(self.openSettings)
        self.run_action.triggered.connect(self.shell.entered)
        self.open_file_action.triggered.connect(self.openScriptFile)
        self.open_in_editor_action.triggered.connect(self.openScriptFileExtEditor)
        self.save_file_action.triggered.connect(self.saveScriptFile)
        self.save_as_file_action.triggered.connect(self.saveAsScriptFile)
        self.help_console_action.triggered.connect(self.openHelpConsole)
        self.help_api_action.triggered.connect(self.openHelpAPI)
        self.help_cookbook_action.triggered.connect(self.openHelpCookbook)
        self.listClassMethod.itemClicked.connect(self.onClickGoToLine)

        if iface is not None:
            self.exit_blocker = ConsoleExitBlocker(self)
            iface.registerApplicationExitBlocker(self.exit_blocker)

    def allowExit(self):
        tab_count = self.tabEditorWidget.count()
        for i in range(tab_count):
            # iterate backwards through tabs, as we may be closing some as we go
            tab_index = tab_count - i - 1
            tab_widget = self.tabEditorWidget.widget(tab_index)
            if tab_widget.isModified():
                ret = QMessageBox.question(
                    self,
                    self.tr("Save {}").format(self.tabEditorWidget.tabText(tab_index)),
                    self.tr(
                        "There are unsaved changes in this script. Do you want to keep those?"
                    ),
                    QMessageBox.StandardButton.Save
                    | QMessageBox.StandardButton.Cancel
                    | QMessageBox.StandardButton.Discard,
                    QMessageBox.StandardButton.Cancel,
                )
                if ret == QMessageBox.StandardButton.Save:
                    tab_widget.save()
                    if tab_widget.isModified():
                        # save failed, treat as cancel
                        return False
                elif ret == QMessageBox.StandardButton.Discard:
                    pass
                else:
                    return False

            self.tabEditorWidget.removeTab(tab_index)

        return True

    def _toggleFind(self):
        self.tabEditorWidget.currentWidget().toggleFindWidget()

    def onClickGoToLine(self, item, column):
        tabEditor = self.tabEditorWidget.currentWidget()
        if item.text(1) == "syntaxError":
            check = tabEditor.syntaxCheck()
            if check and not tabEditor.isReadOnly():
                self.tabEditorWidget.currentWidget().save()
            return
        linenr = int(item.text(1))
        itemName = str(item.text(0))
        charPos = itemName.find(" ")
        if charPos != -1:
            objName = itemName[0:charPos]
        else:
            objName = itemName
        tabEditor.goToLine(str.encode(objName), linenr)

    def toggleEditor(self, checked):
        self.splitterObj.show() if checked else self.splitterObj.hide()
        if not self.tabEditorWidget:
            self.tabEditorWidget.enableToolBarEditor(checked)
            self.tabEditorWidget.restoreTabsOrAddNew()

    def toggleObjectListWidget(self, checked):
        self.listClassMethod.show() if checked else self.listClassMethod.hide()

    def pasteEditor(self):
        self.tabEditorWidget.currentWidget().paste()

    def cutEditor(self):
        self.tabEditorWidget.currentWidget().cut()

    def copyEditor(self):
        self.tabEditorWidget.currentWidget().copy()

    def runScriptEditor(self):
        self.tabEditorWidget.currentWidget().runScriptCode()

    def runSelectedEditor(self):
        self.tabEditorWidget.currentWidget().runSelectedCode()

    def toggleComment(self):
        self.tabEditorWidget.currentWidget().toggleComment()

    def reformatCode(self):
        self.tabEditorWidget.currentWidget().reformatCode()

    def openScriptFileExtEditor(self):
        tabWidget = self.tabEditorWidget.currentWidget()
        tabWidget.open_in_external_editor()

    def openScriptFile(self):
        settings = QgsSettings()
        lastDirPath = settings.value("pythonConsole/lastDirPath", QDir.homePath())
        openFileTr = QCoreApplication.translate("PythonConsole", "Open File")
        fileList, selected_filter = QFileDialog.getOpenFileNames(
            self, openFileTr, lastDirPath, "Script file (*.py)"
        )
        if fileList:
            for pyFile in fileList:
                for i in range(self.tabEditorWidget.count()):
                    tabWidget = self.tabEditorWidget.widget(i)
                    if tabWidget.file_path() == pyFile:
                        self.tabEditorWidget.setCurrentWidget(tabWidget)
                        break
                else:
                    tabName = QFileInfo(pyFile).fileName()
                    self.tabEditorWidget.newTabEditor(tabName, pyFile)

                    lastDirPath = QFileInfo(pyFile).path()
                    settings.setValue("pythonConsole/lastDirPath", pyFile)
                    self.updateTabListScript(pyFile, action="append")

    def saveScriptFile(self):
        tabWidget = self.tabEditorWidget.currentWidget()
        try:
            tabWidget.save()
        except OSError as error:
            msgText = QCoreApplication.translate(
                "PythonConsole", "The file <b>{0}</b> could not be saved. Error: {1}"
            ).format(tabWidget.file_path(), error.strerror)
            self.callWidgetMessageBarEditor(msgText, Qgis.MessageLevel.Critical)

    def saveAsScriptFile(self, index=None):
        tabWidget = self.tabEditorWidget.currentWidget()
        if not index:
            index = self.tabEditorWidget.currentIndex()
        if not tabWidget.file_path():
            fileName = self.tabEditorWidget.tabText(index).replace("*", "")
            fileName = QgsFileUtils.ensureFileNameHasExtension(fileName, ["py"])
            folder = QgsSettings().value("pythonConsole/lastDirPath", QDir.homePath())
            pathFileName = os.path.join(folder, fileName)
            fileNone = True
        else:
            pathFileName = tabWidget.file_path()
            fileNone = False
        saveAsFileTr = QCoreApplication.translate("PythonConsole", "Save File As")
        filename, filter = QFileDialog.getSaveFileName(
            self, saveAsFileTr, pathFileName, "Script file (*.py)"
        )
        if filename:
            filename = QgsFileUtils.ensureFileNameHasExtension(filename, ["py"])

            try:
                tabWidget.save(filename)
            except OSError as error:
                msgText = QCoreApplication.translate(
                    "PythonConsole",
                    "The file <b>{0}</b> could not be saved. Error: {1}",
                ).format(tabWidget.file_path(), error.strerror)
                self.callWidgetMessageBarEditor(msgText, Qgis.MessageLevel.Critical)
                if fileNone:
                    tabWidget.set_file_path(None)
                else:
                    tabWidget.set_file_path(pathFileName)
                return

            if not fileNone:
                self.updateTabListScript(pathFileName, action="remove")

    def openHelpConsole(self):
        QgsHelp.openHelp("plugins/python_console.html")

    def openHelpAPI(self):
        m = re.search(r"^([0-9]+)\.([0-9]+)\.", Qgis.QGIS_VERSION)
        if m:
            QDesktopServices.openUrl(
                QUrl(f"https://qgis.org/pyqgis/{m.group(1)}.{m.group(2)}/")
            )

    def openHelpCookbook(self):
        m = re.search(r"^([0-9]+)\.([0-9]+)\.", Qgis.QGIS_VERSION)
        if m:
            QDesktopServices.openUrl(
                QUrl(
                    f"https://docs.qgis.org/{m.group(1)}.{m.group(2)}/en/docs/pyqgis_developer_cookbook/index.html"
                )
            )

    def openSettings(self):
        iface.showOptionsDialog(iface.mainWindow(), currentPage="consoleOptions")

    def updateSettings(self):
        self.shell.refreshSettingsShell()
        self.shell_output.refreshSettingsOutput()
        self.tabEditorWidget.refreshSettingsEditor()

    def callWidgetMessageBar(self, text):
        self.shell_output.widgetMessageBar(text)

    def callWidgetMessageBarEditor(self, text, level):
        self.tabEditorWidget.showMessage(text, level)

    def updateTabListScript(self, script, action=None):
        if action == "remove":
            self.tabListScript.remove(script)
        elif action == "append":
            if not self.tabListScript:
                self.tabListScript = []
            if script not in self.tabListScript:
                self.tabListScript.append(script)
        else:
            self.tabListScript = []
        QgsSettings().setValue("pythonConsole/tabScripts", self.tabListScript)

    def saveSettingsConsole(self):
        settings = QgsSettings()
        settings.setValue("pythonConsole/splitterConsole", self.splitter.saveState())
        settings.setValue("pythonConsole/splitterObj", self.splitterObj.saveState())
        settings.setValue(
            "pythonConsole/splitterEditor", self.splitterEditor.saveState()
        )

        self.shell.writeHistoryFile()

    def restoreSettingsConsole(self):
        settings = QgsSettings()
        storedTabScripts = settings.value("pythonConsole/tabScripts", [])
        self.tabListScript = storedTabScripts
        self.splitter.restoreState(
            settings.value("pythonConsole/splitterConsole", QByteArray())
        )
        self.splitterEditor.restoreState(
            settings.value("pythonConsole/splitterEditor", QByteArray())
        )
        self.splitterObj.restoreState(
            settings.value("pythonConsole/splitterObj", QByteArray())
        )


if __name__ == "__main__":
    a = QApplication(sys.argv)
    console = PythonConsoleWidget()
    console.show()
    a.exec()



"""
***************************************************************************
    r_li.py
    -------
    Date                 : February 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "February 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

import os
import shutil

from processing.tools.system import getTempFilename, isWindows, mkdir
from qgis.core import QgsProcessingParameterString
from qgis.PyQt.QtCore import QCoreApplication

from grassprovider.grass_utils import GrassUtils

# for MS-Windows users who have MBCS chars in their name:
if os.name == "nt":
    import win32api


def rliPath():
    """Return r.li GRASS user dir"""
    grass_version = GrassUtils.installedVersion().split(".")[0]
    if isWindows():
        homeDir = win32api.GetShortPathName(os.path.expanduser("~"))
        return os.path.join(
            homeDir, "AppData", "Roaming", f"GRASS{grass_version}", "r.li"
        )
    else:
        return os.path.join(os.path.expanduser("~"), f".grass{grass_version}", "r.li")


def removeConfigFile(alg, parameters, context):
    """Remove the r.li user dir config file"""
    configPath = alg.parameterAsString(parameters, "config", context)
    if isWindows():
        command = f"DEL {os.path.join(rliPath(), configPath)}"
    else:
        command = f"rm {os.path.join(rliPath(), configPath)}"
    alg.commands.append(command)


def checkMovingWindow(alg, parameters, context, outputTxt=False):
    """Verify if we have the right parameters"""
    configTxt = alg.parameterAsString(parameters, "config_txt", context)
    config = alg.parameterAsString(parameters, "config", context)
    if configTxt and config:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "You need to set either inline configuration or a configuration file!",
        )

    # Verify that configuration is in moving window
    movingWindow = False
    if configTxt:
        if "MOVINGWINDOW" in configTxt:
            movingWindow = True

    # Read config file:
    if config:
        with open(config) as f:
            for line in f:
                if "MOVINGWINDOW" in line:
                    movingWindow = True

    if not movingWindow and not outputTxt:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            'Your configuration needs to be a "moving window" configuration!',
        )

    if movingWindow and outputTxt:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            'Your configuration needs to be a non "moving window" configuration!',
        )

    return True, None


def configFile(alg, parameters, context, feedback, outputTxt=False):
    """Handle inline configuration
    :param parameters:
    """
    # Where is the GRASS user directory ?
    user_grass_path = rliPath()
    if not os.path.isdir(user_grass_path):
        mkdir(user_grass_path)
    if not os.path.isdir(os.path.join(user_grass_path, "output")):
        mkdir(os.path.join(user_grass_path, "output"))

    # If we have a configuration file, we need to copy it into user dir
    if parameters["config"]:
        fileName = alg.parameterAsString(parameters, "config", context)
        configFilePath = os.path.join(user_grass_path, os.path.basename(fileName))
        # Copy the file
        shutil.copy(parameters["config"], configFilePath)
        # Change the parameter value
        parameters["config"] = os.path.basename(configFilePath)
    # Handle inline configuration
    elif parameters["config_txt"]:
        # Creates a temporary txt file in user r.li directory
        tempConfig = os.path.basename(getTempFilename(context=context))
        configFilePath = os.path.join(user_grass_path, tempConfig)
        # Inject rules into temporary txt file
        with open(configFilePath, "w") as f:
            f.write(alg.parameterAsString(parameters, "config_txt", context))
            f.write("\n")

        # Use temporary file as rules file
        parameters["config"] = os.path.basename(configFilePath)
        alg.removeParameter("config_txt")

    # For ascii output, we need a virtual output
    if outputTxt:
        param = QgsProcessingParameterString(
            "output",
            "virtual output",
            "a" + os.path.basename(getTempFilename(context=context)),
            False,
            False,
        )
        alg.addParameter(param)

    alg.processCommand(parameters, context, feedback, outputTxt)

    # Remove Config file:
    removeConfigFile(alg, parameters, context)


def moveOutputTxtFile(alg, parameters, context):
    # Find output file name:
    txtPath = alg.parameterAsString(parameters, "output_txt", context)
    user_grass_path = rliPath()

    output = os.path.join(
        user_grass_path, "output", alg.parameterAsString(parameters, "output", context)
    )
    # move the file
    if isWindows():
        command = f"MOVE /Y {output} {txtPath}"
    else:
        command = f"mv -f {output} {txtPath}"
    alg.commands.append(command)



"""
***************************************************************************
    r_horizon.py
    ------------
    Date                 : September 2017
    Copyright            : (C) 2017 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "September 2017"
__copyright__ = "(C) 2017, Médéric Ribreux"

import math
import os

from qgis.PyQt.QtCore import QCoreApplication


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    start = alg.parameterAsDouble(parameters, "start", context)
    end = alg.parameterAsDouble(parameters, "end", context)
    step = alg.parameterAsDouble(parameters, "step", context)

    if start >= end:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "The start position must be inferior to the end position!",
        )
    if step == 0.0:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt", "The step must be greater than zero!"
        )
    return True, None


def processOutputs(alg, parameters, context, feedback):
    # Inspired from GRASS implementation
    def getNumberDecimals(number):
        """Return the number of decimals of a number (int or float)"""
        if int(number) == number:
            return 0
        return len(str(number).split(".")[-1])

    def doubleToBaseName(number, nDecimals):
        """
        Format filename, according to GRASS implementation,
        based on provided number and number of decimals
        """
        number += 0.0001
        # adapted from GRASS https://github.com/OSGeo/grass/blob/6253da1bd6ce48d23419e99e8b503edf46178490/lib/gis/basename.c#L97-L101
        if nDecimals == 0:
            return f"{int(number):03}"
        int_part = int(number)
        dec_part = int((number - int_part) * pow(10, nDecimals))
        return f"{int_part:03}_{str(dec_part).rjust(nDecimals, '0')}"

    # There will be as many outputs as the difference between start and end divided by steps
    start = alg.parameterAsDouble(parameters, "start", context)
    end = alg.parameterAsDouble(parameters, "end", context)
    step = alg.parameterAsDouble(parameters, "step", context)
    direction = alg.parameterAsDouble(parameters, "direction", context)

    first_rad = math.radians(start + direction)
    nDecimals = getNumberDecimals(step)
    dfr_rad = math.radians(step)
    arrayNumInt = int((end - start) / abs(step))

    directory = alg.parameterAsString(parameters, "output", context)
    # Needed if output to a temporary directory
    os.makedirs(directory, exist_ok=True)
    for k in range(arrayNumInt):
        angle_deg = math.degrees(first_rad + dfr_rad * k)
        baseName = doubleToBaseName(angle_deg, nDecimals)
        grassName = f"output{alg.uniqueSuffix}_{baseName}"
        fileName = f"{os.path.join(directory, baseName)}.tif"
        alg.exportRasterLayer(grassName, fileName)



"""
***************************************************************************
    sql_dictionary.py
    ---------------------
    Date                 : April 2012
    Copyright            : (C) 2012 by Giuseppe Sucameli
    Email                : brush dot tyler at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Giuseppe Sucameli"
__date__ = "April 2012"
__copyright__ = "(C) 2012, Giuseppe Sucameli"

# GENERIC SQL DICTIONARY

# keywords
keywords = [
    "action",
    "add",
    "after",
    "all",
    "alter",
    "analyze",
    "and",
    "as",
    "asc",
    "before",
    "begin",
    "between",
    "by",
    "cascade",
    "case",
    "cast",
    "check",
    "collate",
    "column",
    "commit",
    "constraint",
    "create",
    "cross",
    "current_date",
    "current_time",
    "current_timestamp",
    "default",
    "deferrable",
    "deferred",
    "delete",
    "desc",
    "distinct",
    "drop",
    "each",
    "else",
    "end",
    "escape",
    "except",
    "exists",
    "for",
    "foreign",
    "from",
    "full",
    "group",
    "having",
    "ignore",
    "immediate",
    "in",
    "initially",
    "inner",
    "insert",
    "intersect",
    "into",
    "is",
    "isnull",
    "join",
    "key",
    "left",
    "like",
    "limit",
    "match",
    "natural",
    "no",
    "not",
    "notnull",
    "null",
    "of",
    "offset",
    "on",
    "or",
    "order",
    "outer",
    "primary",
    "references",
    "release",
    "restrict",
    "right",
    "rollback",
    "row",
    "savepoint",
    "select",
    "set",
    "table",
    "temporary",
    "then",
    "to",
    "transaction",
    "trigger",
    "union",
    "unique",
    "update",
    "using",
    "values",
    "view",
    "when",
    "where",
]

# functions
functions = [
    "abs",
    "changes",
    "coalesce",
    "glob",
    "ifnull",
    "hex",
    "last_insert_rowid",
    "length",
    "like",
    "lower",
    "ltrim",
    "max",
    "min",
    "nullif",
    "quote",
    "random",
    "randomblob",
    "replace",
    "round",
    "rtrim",
    "soundex",
    "total_change",
    "trim",
    "typeof",
    "upper",
    "zeroblob",
    "date",
    "datetime",
    "julianday",
    "strftime",
    "avg",
    "count",
    "group_concat",
    "sum",
    "total",
]

# constants
constants = ["null", "false", "true"]


def getSqlDictionary():
    return {
        "keyword": list(keywords),
        "constant": list(constants),
        "function": list(functions),
    }



"""
/***************************************************************************
Name                 : Virtual layers plugin for DB Manager
Date                 : December 2015
copyright            : (C) 2015 by Hugo Mercier
email                : hugo dot mercier at oslandia dot com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import (
    QgsTask,
    QgsVectorLayer,
    QgsVirtualLayerDefinition,
    QgsVirtualLayerTask,
    QgsWkbTypes,
)
from qgis.PyQt.QtCore import QElapsedTimer, QTemporaryFile

from ..data_model import (
    BaseTableModel,
    SqlResultModelAsync,
    SqlResultModelTask,
    TableDataModel,
)
from ..plugin import BaseError, DbError
from .connector import VLayerRegistry, getQueryGeometryName
from .plugin import LVectorTable


class LTableDataModel(TableDataModel):
    def __init__(self, table, parent=None):
        TableDataModel.__init__(self, table, parent)

        self.layer = None

        if isinstance(table, LVectorTable):
            self.layer = VLayerRegistry.instance().getLayer(table.name)
        else:
            self.layer = VLayerRegistry.instance().getLayer(table)

        if not self.layer:
            return
        # populate self.resdata
        self.resdata = []
        for f in self.layer.getFeatures():
            a = f.attributes()
            # add the geometry type
            if f.hasGeometry():
                a.append(QgsWkbTypes.displayString(f.geometry().wkbType()))
            else:
                a.append("None")
            self.resdata.append(a)

        self.fetchedFrom = 0
        self.fetchedCount = len(self.resdata)

    def rowCount(self, index=None):
        if self.layer:
            return self.layer.featureCount()
        return 0


class LSqlResultModelTask(SqlResultModelTask):
    def __init__(self, db, sql, parent):
        super().__init__(db, sql, parent)

        tf = QTemporaryFile()
        tf.open()
        path = tf.fileName()
        tf.close()

        df = QgsVirtualLayerDefinition()
        df.setFilePath(path)
        df.setQuery(sql)

        self.subtask = QgsVirtualLayerTask(df)
        self.addSubTask(
            self.subtask, [], QgsTask.SubTaskDependency.ParentDependsOnSubTask
        )

    def run(self):
        try:
            path = self.subtask.definition().filePath()
            sql = self.subtask.definition().query()
            self.model = LSqlResultModel(self.db, sql, None, self.subtask.layer(), path)
        except Exception as e:
            self.error = BaseError(str(e))
            return False
        return True

    def cancel(self):
        SqlResultModelTask.cancel(self)


class LSqlResultModelAsync(SqlResultModelAsync):
    def __init__(self, db, sql, parent=None):
        super().__init__()

        self.task = LSqlResultModelTask(db, sql, parent)
        self.task.taskCompleted.connect(self.modelDone)
        self.task.taskTerminated.connect(self.modelDone)

    def modelDone(self):
        self.status = self.task.status
        self.model = self.task.model
        if self.task.subtask.exceptionText():
            self.error = BaseError(self.task.subtask.exceptionText())
        self.done.emit()


class LSqlResultModel(BaseTableModel):
    def __init__(self, db, sql, parent=None, layer=None, path=None):
        t = QElapsedTimer()
        t.start()

        if not layer:
            tf = QTemporaryFile()
            tf.open()
            path = tf.fileName()
            tf.close()

            df = QgsVirtualLayerDefinition()
            df.setFilePath(path)
            df.setQuery(sql)
            layer = QgsVectorLayer(df.toString(), "vv", "virtual")
            self._secs = t.elapsed() / 1000.0

        data = []
        header = []

        if not layer.isValid():
            raise DbError(layer.dataProvider().error().summary(), sql)
        else:
            header = [f.name() for f in layer.fields()]
            has_geometry = False
            if layer.geometryType() != QgsWkbTypes.GeometryType.NullGeometry:
                gn = getQueryGeometryName(path)
                if gn:
                    has_geometry = True
                    header += [gn]

            for f in layer.getFeatures():
                a = f.attributes()
                if has_geometry:
                    if f.hasGeometry():
                        a += [f.geometry().asWkt()]
                    else:
                        a += [None]
                data += [a]

        self._secs = 0
        self._affectedRows = len(data)

        BaseTableModel.__init__(self, header, data, parent)

    def secs(self):
        return self._secs

    def affectedRows(self):
        return self._affectedRows



"""
***************************************************************************
    r_drain.py
    ----------
    Date                 : February 2016
    Copyright            : (C) 2016 by Médéric Ribreux
    Email                : medspx at medspx dot fr
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Médéric Ribreux"
__date__ = "February 2016"
__copyright__ = "(C) 2016, Médéric Ribreux"

from qgis.PyQt.QtCore import QCoreApplication


def checkParameterValuesBeforeExecuting(alg, parameters, context):
    """Verify if we have the right parameters"""
    # start_coordinates and start_points are mutually exclusives
    if alg.parameterAsString(
        parameters, "start_coordinates", context
    ) and alg.parameterAsVectorLayer(parameters, "start_points", context):
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "You need to set either start coordinates OR a start points vector layer!",
        )

    # You need to set at least one parameter
    if not alg.parameterAsString(
        parameters, "start_coordinates", context
    ) and not alg.parameterAsVectorLayer(parameters, "start_points", context):
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt",
            "You need to set either start coordinates OR a start points vector layer!",
        )

    paramscore = [
        f for f in ["-c", "-a", "-n"] if alg.parameterAsBoolean(parameters, f, context)
    ]
    if len(paramscore) > 1:
        return False, QCoreApplication.translate(
            "GrassAlgorithmExt", "-c, -a, -n parameters are mutually exclusive!"
        )
    return True, None



"""
/***************************************************************************
Name                 : DB Manager
Description          : Database manager plugin for QGIS
Date                 : May 23, 2011
copyright            : (C) 2011 by Giuseppe Sucameli
email                : brush.tyler@gmail.com

 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/
"""

from qgis.core import QgsTask
from qgis.PyQt.QtCore import (
    QAbstractTableModel,
    QElapsedTimer,
    QObject,
    QRegularExpression,
    Qt,
    pyqtSignal,
)
from qgis.PyQt.QtGui import QFont, QStandardItem, QStandardItemModel
from qgis.PyQt.QtWidgets import QApplication

from .plugin import BaseError, DbError


class BaseTableModel(QAbstractTableModel):
    def __init__(self, header=None, data=None, parent=None):
        QAbstractTableModel.__init__(self, parent)
        self._header = header if header else []
        self.resdata = data if data else []

    def headerToString(self, sep="\t"):
        header = self._header
        return sep.join(header)

    def rowToString(self, row, sep="\t"):
        return sep.join(
            str(self.getData(row, col)) for col in range(self.columnCount())
        )

    def getData(self, row, col):
        return self.resdata[row][col]

    def columnNames(self):
        return list(self._header)

    def rowCount(self, parent=None):
        return len(self.resdata)

    def columnCount(self, parent=None):
        return len(self._header)

    def data(self, index, role):
        if role not in [
            Qt.ItemDataRole.DisplayRole,
            Qt.ItemDataRole.EditRole,
            Qt.ItemDataRole.FontRole,
        ]:
            return None

        val = self.getData(index.row(), index.column())

        if role == Qt.ItemDataRole.EditRole:
            return val

        if role == Qt.ItemDataRole.FontRole:  # draw NULL in italic
            if val is not None:
                return None
            f = QFont()
            f.setItalic(True)
            return f

        if val is None:
            return "NULL"
        elif isinstance(val, memoryview):
            # hide binary data
            return None
        elif isinstance(val, str) and len(val) > 300:
            # too much data to display, elide the string
            val = val[:300]
        try:
            return str(val)  # convert to Unicode
        except UnicodeDecodeError:
            return str(
                val, "utf-8", "replace"
            )  # convert from utf8 and replace errors (if any)

    def headerData(self, section, orientation, role):
        if role != Qt.ItemDataRole.DisplayRole:
            return None

        if orientation == Qt.Orientation.Vertical:
            # header for a row
            return section + 1
        else:
            # header for a column
            return self._header[section]


class TableDataModel(BaseTableModel):
    def __init__(self, table, parent=None):
        self.db = table.database().connector
        self.table = table

        fieldNames = [x.name for x in table.fields()]
        BaseTableModel.__init__(self, fieldNames, None, parent)

        # get table fields
        self.fields = []
        for fld in table.fields():
            self.fields.append(self._sanitizeTableField(fld))

        self.fetchedCount = 201
        self.fetchedFrom = (
            -self.fetchedCount - 1
        )  # so the first call to getData will exec fetchMoreData(0)

    def _sanitizeTableField(self, field):
        """quote column names to avoid some problems (e.g. columns with upper case)"""
        return self.db.quoteId(field)

    def getData(self, row, col):
        if row < self.fetchedFrom or row >= self.fetchedFrom + self.fetchedCount:
            margin = self.fetchedCount / 2
            start = int(
                self.rowCount() - margin
                if row + margin >= self.rowCount()
                else row - margin
            )
            if start < 0:
                start = 0
            self.fetchMoreData(start)
        return self.resdata[row - self.fetchedFrom][col]

    def fetchMoreData(self, row_start):
        pass

    def rowCount(self, index=None):
        # case for tables with no columns ... any reason to use them? :-)
        return (
            self.table.rowCount
            if self.table.rowCount is not None and self.columnCount(index) > 0
            else 0
        )


class SqlResultModelAsync(QObject):
    done = pyqtSignal()

    def __init__(self):
        super().__init__()
        self.error = BaseError("")
        self.status = None
        self.model = None
        self.task = None
        self.canceled = False

    def cancel(self):
        self.canceled = True
        if self.task:
            self.task.cancel()

    def modelDone(self):
        if self.task:
            self.status = self.task.status
            self.model = self.task.model
            self.error = self.task.error

        self.done.emit()


class SqlResultModelTask(QgsTask):
    def __init__(self, db, sql, parent):
        super().__init__(
            description=QApplication.translate("DBManagerPlugin", "Executing SQL")
        )
        self.db = db
        self.sql = sql
        self.parent = parent
        self.error = BaseError("")
        self.model = None


class SqlResultModel(BaseTableModel):
    def __init__(self, db, sql, parent=None):
        self.db = db.connector

        t = QElapsedTimer()
        t.start()
        c = self.db._execute(None, sql)

        self._affectedRows = 0
        data = []
        header = self.db._get_cursor_columns(c)
        if header is None:
            header = []

        try:
            if len(header) > 0:
                data = self.db._fetchall(c)
            self._affectedRows = len(data)
        except DbError:
            # nothing to fetch!
            data = []
            header = []

        super().__init__(header, data, parent)

        # commit before closing the cursor to make sure that the changes are stored
        self.db._commit()
        c.close()
        self._secs = t.elapsed() / 1000.0
        del c
        del t

    def secs(self):
        return self._secs

    def affectedRows(self):
        return self._affectedRows


class SimpleTableModel(QStandardItemModel):
    def __init__(self, header, editable=False, parent=None):
        self.header = header
        self.editable = editable
        QStandardItemModel.__init__(self, 0, len(self.header), parent)

    def rowFromData(self, data):
        row = []
        for c in data:
            item = QStandardItem(str(c))
            item.setFlags(
                (item.flags() | Qt.ItemFlag.ItemIsEditable)
                if self.editable
                else (item.flags() & ~Qt.ItemFlag.ItemIsEditable)
            )
            row.append(item)
        return row

    def headerData(self, section, orientation, role):
        if (
            orientation == Qt.Orientation.Horizontal
            and role == Qt.ItemDataRole.DisplayRole
        ):
            return self.header[section]
        return None

    def _getNewObject(self):
        pass

    def getObject(self, row):
        return self._getNewObject()

    def getObjectIter(self):
        for row in range(self.rowCount()):
            yield self.getObject(row)


class TableFieldsModel(SimpleTableModel):
    def __init__(self, parent, editable=False):
        SimpleTableModel.__init__(
            self, ["Name", "Type", "Null", "Default", "Comment"], editable, parent
        )

    def headerData(self, section, orientation, role):
        if (
            orientation == Qt.Orientation.Vertical
            and role == Qt.ItemDataRole.DisplayRole
        ):
            return section + 1
        return SimpleTableModel.headerData(self, section, orientation, role)

    def flags(self, index):
        flags = SimpleTableModel.flags(self, index)
        if (
            index.column() == 2 and flags & Qt.ItemFlag.ItemIsEditable
        ):  # set Null column as checkable instead of editable
            flags = (
                flags & ~Qt.ItemFlag.ItemIsEditable | Qt.ItemFlag.ItemIsUserCheckable
            )
        return flags

    def append(self, fld):
        data = [
            fld.name,
            fld.type2String(),
            not fld.notNull,
            fld.default2String(),
            fld.getComment(),
        ]
        self.appendRow(self.rowFromData(data))
        row = self.rowCount() - 1
        self.setData(self.index(row, 0), fld, Qt.ItemDataRole.UserRole)
        self.setData(self.index(row, 1), fld.primaryKey, Qt.ItemDataRole.UserRole)
        self.setData(self.index(row, 2), None, Qt.ItemDataRole.DisplayRole)
        self.setData(
            self.index(row, 2),
            Qt.CheckState.Unchecked if fld.notNull else Qt.CheckState.Checked,
            Qt.ItemDataRole.CheckStateRole,
        )

    def _getNewObject(self):
        from .plugin import TableField

        return TableField(None)

    def getObject(self, row):
        val = self.data(self.index(row, 0), Qt.ItemDataRole.UserRole)
        fld = val if val is not None else self._getNewObject()
        fld.name = self.data(self.index(row, 0)) or ""
        typestr = self.data(self.index(row, 1)) or ""
        regex = QRegularExpression(r"([^\(]+)\(([^\)]+)\)")
        match = regex.match(typestr)
        if match.hasMatch():
            fld.dataType = match.captured(1).strip()
            fld.modifier = match.captured(2).strip()
        else:
            fld.modifier = None
            fld.dataType = typestr

        fld.notNull = (
            self.data(self.index(row, 2), Qt.ItemDataRole.CheckStateRole)
            == Qt.CheckState.Unchecked
        )
        fld.primaryKey = self.data(self.index(row, 1), Qt.ItemDataRole.UserRole)
        fld.comment = self.data(self.index(row, 4))
        return fld

    def getFields(self):
        return [fld for fld in self.getObjectIter()]


class TableConstraintsModel(SimpleTableModel):
    def __init__(self, parent, editable=False):
        SimpleTableModel.__init__(
            self,
            [
                QApplication.translate("DBManagerPlugin", "Name"),
                QApplication.translate("DBManagerPlugin", "Type"),
                QApplication.translate("DBManagerPlugin", "Column(s)"),
            ],
            editable,
            parent,
        )

    def append(self, constr):
        field_names = [str(k_v[1].name) for k_v in iter(list(constr.fields().items()))]
        data = [constr.name, constr.type2String(), ", ".join(field_names)]
        self.appendRow(self.rowFromData(data))
        row = self.rowCount() - 1
        self.setData(self.index(row, 0), constr, Qt.ItemDataRole.UserRole)
        self.setData(self.index(row, 1), constr.type, Qt.ItemDataRole.UserRole)
        self.setData(self.index(row, 2), constr.columns, Qt.ItemDataRole.UserRole)

    def _getNewObject(self):
        from .plugin import TableConstraint

        return TableConstraint(None)

    def getObject(self, row):
        constr = self.data(self.index(row, 0), Qt.ItemDataRole.UserRole)
        if not constr:
            constr = self._getNewObject()
        constr.name = self.data(self.index(row, 0)) or ""
        constr.type = self.data(self.index(row, 1), Qt.ItemDataRole.UserRole)
        constr.columns = self.data(self.index(row, 2), Qt.ItemDataRole.UserRole)
        return constr

    def getConstraints(self):
        return [constr for constr in self.getObjectIter()]


class TableIndexesModel(SimpleTableModel):
    def __init__(self, parent, editable=False):
        SimpleTableModel.__init__(
            self,
            [
                QApplication.translate("DBManagerPlugin", "Name"),
                QApplication.translate("DBManagerPlugin", "Column(s)"),
            ],
            editable,
            parent,
        )

    def append(self, idx):
        field_names = [str(k_v1[1].name) for k_v1 in iter(list(idx.fields().items()))]
        data = [idx.name, ", ".join(field_names)]
        self.appendRow(self.rowFromData(data))
        row = self.rowCount() - 1
        self.setData(self.index(row, 0), idx, Qt.ItemDataRole.UserRole)
        self.setData(self.index(row, 1), idx.columns, Qt.ItemDataRole.UserRole)

    def _getNewObject(self):
        from .plugin import TableIndex

        return TableIndex(None)

    def getObject(self, row):
        idx = self.data(self.index(row, 0), Qt.ItemDataRole.UserRole)
        if not idx:
            idx = self._getNewObject()
        idx.name = self.data(self.index(row, 0))
        idx.columns = self.data(self.index(row, 1), Qt.ItemDataRole.UserRole)
        return idx

    def getIndexes(self):
        return [idx for idx in self.getObjectIter()]



"""
***************************************************************************
    Datasources2Vrt.py
    ---------------------
    Date                 : May 2015
    Copyright            : (C) 2015 by Luigi Pirelli
    Email                : luipir at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Luigi Pirelli"
__date__ = "May 2015"
__copyright__ = "(C) 2015, Luigi Pirelli"

import html
import pathlib

from qgis.core import (
    QgsProcessing,
    QgsProcessingException,
    QgsProcessingOutputString,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameters,
    QgsProcessingParameterVectorDestination,
)
from qgis.PyQt.QtCore import QCoreApplication

from processing.algs.gdal.GdalAlgorithm import GdalAlgorithm
from processing.algs.gdal.GdalUtils import GdalUtils


class Datasources2Vrt(GdalAlgorithm):
    INPUT = "INPUT"
    UNIONED = "UNIONED"
    OUTPUT = "OUTPUT"
    VRT_STRING = "VRT_STRING"

    def createCustomParametersWidget(self, parent):
        return None

    def group(self):
        return self.tr("Vector miscellaneous")

    def groupId(self):
        return "vectormiscellaneous"

    def name(self):
        return "buildvirtualvector"

    def displayName(self):
        return self.tr("Build virtual vector")

    def tags(self):
        return ["ogr", "gdal", "vrt", "create"]

    def shortHelpString(self):
        return self.tr(
            "This algorithm creates a virtual layer that contains a set of vector layers."
        )

    def __init__(self):
        super().__init__()

    def initAlgorithm(self, config=None):
        self.addParameter(
            QgsProcessingParameterMultipleLayers(
                self.INPUT,
                self.tr("Input datasources"),
                QgsProcessing.SourceType.TypeVector,
            )
        )
        self.addParameter(
            QgsProcessingParameterBoolean(
                self.UNIONED, self.tr('Create "unioned" VRT'), defaultValue=False
            )
        )

        class ParameterVectorVrtDestination(QgsProcessingParameterVectorDestination):
            def __init__(self, name, description):
                super().__init__(name, description)

            def clone(self):
                copy = ParameterVectorVrtDestination(self.name(), self.description())
                return copy

            def defaultFileExtension(self):
                return "vrt"

            def createFileFilter(self):
                return "{} (*.vrt *.VRT)".format(
                    QCoreApplication.translate("GdalAlgorithm", "VRT files")
                )

            def supportedOutputRasterLayerExtensions(self):
                return ["vrt"]

            def isSupportedOutputValue(self, value, context):
                output_path = QgsProcessingParameters.parameterAsOutputLayer(
                    self, value, context, testOnly=True
                )
                if pathlib.Path(output_path).suffix.lower() != ".vrt":
                    return False, QCoreApplication.translate(
                        "GdalAlgorithm", "Output filename must use a .vrt extension"
                    )
                return True, ""

        self.addParameter(
            ParameterVectorVrtDestination(self.OUTPUT, self.tr("Virtual vector"))
        )
        self.addOutput(
            QgsProcessingOutputString(self.VRT_STRING, self.tr("Virtual string"))
        )

    def processAlgorithm(self, parameters, context, feedback):
        input_layers = self.parameterAsLayerList(parameters, self.INPUT, context)
        unioned = self.parameterAsBoolean(parameters, self.UNIONED, context)
        vrtPath = self.parameterAsOutputLayer(parameters, self.OUTPUT, context)

        vrt = "<OGRVRTDataSource>"
        if unioned:
            vrt += '<OGRVRTUnionLayer name="UnionedLayer">'

        total = 100.0 / len(input_layers) if input_layers else 0
        for current, layer in enumerate(input_layers):
            if feedback.isCanceled():
                break

            connection_details = GdalUtils.gdal_connection_details_from_layer(layer)
            basePath = connection_details.connection_string
            layerName = GdalUtils.ogrLayerName(layer.source())

            vrt += f'<OGRVRTLayer name="{html.escape(layerName, True)}">'
            vrt += f"<SrcDataSource>{html.escape(basePath, True)}</SrcDataSource>"
            vrt += f"<SrcLayer>{html.escape(layerName, True)}</SrcLayer>"
            vrt += "</OGRVRTLayer>"

            feedback.setProgress(int(current * total))

        if unioned:
            vrt += "</OGRVRTUnionLayer>"
        vrt += "</OGRVRTDataSource>"

        with open(vrtPath, "w", encoding="utf-8") as f:
            f.write(vrt)

        return {self.OUTPUT: vrtPath, self.VRT_STRING: vrt}

    def commandName(self):
        return ""



"""
***************************************************************************
    MultilineTextPanel.py
    ---------------------
    Date                 : January 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

__author__ = "Victor Olaya"
__date__ = "January 2013"
__copyright__ = "(C) 2013, Victor Olaya"

from qgis.PyQt.QtWidgets import (
    QComboBox,
    QPlainTextEdit,
    QSizePolicy,
    QVBoxLayout,
    QWidget,
)


class MultilineTextPanel(QWidget):
    USE_TEXT = 0

    def __init__(self, options, parent=None):
        super().__init__(parent)
        self.options = options
        self.verticalLayout = QVBoxLayout(self)
        self.verticalLayout.setSpacing(2)
        self.verticalLayout.setMargin(0)
        self.combo = QComboBox()
        self.combo.addItem(self.tr("[Use text below]"))
        for option in options:
            self.combo.addItem(option[0], option[1])
        self.combo.setSizePolicy(
            QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding
        )
        self.verticalLayout.addWidget(self.combo)
        self.textBox = QPlainTextEdit()
        self.verticalLayout.addWidget(self.textBox)
        self.setLayout(self.verticalLayout)

    def setText(self, text):
        self.textBox.setPlainText(text)

    def getOption(self):
        return self.combo.currentIndex()

    def getValue(self):
        if self.combo.currentIndex() == 0:
            return str(self.textBox.toPlainText())
        else:
            return self.combo.currentData()

    def setValue(self, value):
        items = [self.combo.itemData(i) for i in range(1, self.combo.count())]
        for idx, item in enumerate(items):
            if item == value:
                self.combo.setCurrentIndex(idx)
                return
        self.combo.setCurrentIndex(0)
        if value:
            self.textBox.setPlainText(value)



"""
***************************************************************************
    TestTools.py
    ---------------------
    Date                 : February 2013
    Copyright            : (C) 2013 by Victor Olaya
    Email                : volayaf at gmail dot com
***************************************************************************
*                                                                         *
*   This program is free software; you can redistribute it and/or modify  *
*   it under the terms of the GNU General Public License as published by  *
*   the Free Software Foundation; either version 2 of the License, or     *
*   (at your option) any later version.                                   *
*                                                                         *
***************************************************************************
"""

import ast
import hashlib
import os
import posixpath
import re

import yaml
from numpy import nan_to_num
from osgeo import gdal
from osgeo.gdalconst import GA_ReadOnly
from qgis.core import (
    QgsApplication,
    QgsProcessing,
    QgsProcessingParameterBand,
    QgsProcessingParameterBoolean,
    QgsProcessingParameterDefinition,
    QgsProcessingParameterDistance,
    QgsProcessingParameterDuration,
    QgsProcessingParameterEnum,
    QgsProcessingParameterFeatureSink,
    QgsProcessingParameterFeatureSource,
    QgsProcessingParameterFile,
    QgsProcessingParameterFileDestination,
    QgsProcessingParameterMultipleLayers,
    QgsProcessingParameterNumber,
    QgsProcessingParameterRasterDestination,
    QgsProcessingParameterRasterLayer,
    QgsProcessingParameterString,
    QgsProcessingParameterVectorDestination,
    QgsProcessingParameterVectorLayer,
    QgsProperty,
)
from qgis.PyQt.QtCore import QCoreApplication, QMetaObject
from qgis.PyQt.QtWidgets import QDialog, QMessageBox, QTextEdit, QVBoxLayout

gdal.UseExceptions()


def extractSchemaPath(filepath):
    """
    Tries to find where the file is relative to the QGIS source code directory.
    If it is already placed in the processing or QGIS testdata directory it will
    return an appropriate schema and relative filepath

    Args:
        filepath: The path of the file to examine

    Returns:
        A tuple (schema, relative_file_path) where the schema is 'qgs' or 'proc'
        if we can assume that the file is in this testdata directory.
    """
    parts = []
    schema = None
    localpath = ""
    path = filepath
    part = True

    while part and filepath:
        (path, part) = os.path.split(path)
        if part == "testdata" and not localpath:
            localparts = parts
            localparts.reverse()
            # we always want posix style paths here
            localpath = posixpath.join(*localparts)

        parts.append(part)

    parts.reverse()

    try:
        testsindex = parts.index("tests")
    except ValueError:
        return "", filepath

    if parts[testsindex - 1] == "processing":
        schema = "proc"

    return schema, localpath


def parseParameters(command):
    """
    Parse alg string to grab parameters value.
    Can handle quotes and comma.
    """
    pos = 0
    exp = re.compile(r"""(['"]?)(.*?)\1(,|$)""")
    while True:
        m = exp.search(command, pos)
        result = m.group(2)
        separator = m.group(3)

        # Handle special values:
        if result == "None":
            result = None
        elif result.lower() == str(True).lower():
            result = True
        elif result.lower() == str(False).lower():
            result = False

        yield result

        if not separator:
            break

        pos = m.end(0)


def splitAlgIdAndParameters(command: str) -> tuple[str, dict]:
    """
    Extracts the algorithm ID and input parameter list from a processing runalg command
    """
    exp = re.compile(r"""['"](.*?)['"]\s*,\s*(.*)""")
    m = exp.search(command[len("processing.run(") : -1])
    alg_id = m.group(1)
    param_string = m.group(2)

    # replace QgsCoordinateReferenceSystem('EPSG:4325') with just string value
    exp = re.compile(r"""QgsCoordinateReferenceSystem\((['"].*?['"])\)""")
    param_string = exp.sub("\\1", param_string)
    exp = re.compile(r"""(QgsProperty.from[a-zA-Z]+\(.*?\))([,}])""")
    param_string = exp.sub("'''\\1'''\\2", param_string)

    return alg_id, ast.literal_eval(param_string)


def createTest(text):
    definition = {}

    alg_id, parameters = splitAlgIdAndParameters(text)

    alg = QgsApplication.processingRegistry().createAlgorithmById(alg_id)

    definition["name"] = f"Test ({alg_id})"
    definition["algorithm"] = alg_id

    params = {}
    results = {}

    i = 0
    for param in alg.parameterDefinitions():
        if (
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden
            or param.isDestination()
        ):
            continue

        if not param.name() in parameters:
            continue

        i += 1
        token = parameters[param.name()]
        # Handle empty parameters that are optionals
        if (
            param.flags() & QgsProcessingParameterDefinition.Flag.FlagOptional
            and token is None
        ):
            continue

        if isinstance(token, str) and token.startswith("QgsProperty"):
            # dynamic property
            match = re.match(
                r"""QgsProperty.fromExpression\(\s*['"](.*)['"]\s*\)""", token
            )
            if match:
                p = {"type": "property", "expression": match.group(1)}
                params[param.name()] = p
            else:
                assert False, "Only expression based properties currently handled"
        elif isinstance(
            param,
            (QgsProcessingParameterVectorLayer, QgsProcessingParameterFeatureSource),
        ):
            schema, filepath = extractSchemaPath(token)
            p = {"type": "vector", "name": filepath}
            if not schema:
                p["location"] = (
                    "[The source data is not in the testdata directory. Please use data in the processing/tests/testdata folder.]"
                )

            params[param.name()] = p
        elif isinstance(param, QgsProcessingParameterRasterLayer):
            schema, filepath = extractSchemaPath(token)
            p = {"type": "raster", "name": filepath}
            if not schema:
                p["location"] = (
                    "[The source data is not in the testdata directory. Please use data in the processing/tests/testdata folder.]"
                )

            params[param.name()] = p
        elif isinstance(param, QgsProcessingParameterMultipleLayers):
            multiparams = token
            newparam = []

            # Handle datatype detection
            dataType = param.layerType()
            if dataType in [
                QgsProcessing.SourceType.TypeVectorAnyGeometry,
                QgsProcessing.SourceType.TypeVectorPoint,
                QgsProcessing.SourceType.TypeVectorLine,
                QgsProcessing.SourceType.TypeVectorPolygon,
                QgsProcessing.SourceType.TypeVector,
            ]:
                dataType = "vector"
            else:
                dataType = "raster"

            schema = None
            for mp in multiparams:
                schema, filepath = extractSchemaPath(mp)
                newparam.append({"type": dataType, "name": filepath})
            p = {"type": "multi", "params": newparam}
            if not schema:
                p["location"] = (
                    "[The source data is not in the testdata directory. Please use data in the processing/tests/testdata folder.]"
                )

            params[param.name()] = p
        elif isinstance(param, QgsProcessingParameterFile):
            schema, filepath = extractSchemaPath(token)
            p = {"type": "file", "name": filepath}
            if not schema:
                p["location"] = (
                    "[The source data is not in the testdata directory. Please use data in the processing/tests/testdata folder.]"
                )

            params[param.name()] = p
        elif isinstance(param, QgsProcessingParameterString):
            params[param.name()] = token
        elif isinstance(param, QgsProcessingParameterBoolean):
            params[param.name()] = token
        elif isinstance(
            param, (QgsProcessingParameterNumber, QgsProcessingParameterDistance)
        ):
            if param.dataType() == QgsProcessingParameterNumber.Type.Integer:
                params[param.name()] = int(token)
            else:
                params[param.name()] = float(token)
        elif isinstance(param, QgsProcessingParameterEnum):
            if isinstance(token, list):
                params[param.name()] = [int(t) for t in token]
            else:
                params[param.name()] = int(token)
        elif isinstance(param, QgsProcessingParameterBand):
            params[param.name()] = int(token)
        elif token:
            if token[0] == '"':
                token = token[1:]
            if token[-1] == '"':
                token = token[:-1]
            params[param.name()] = token

    definition["params"] = params

    for i, out in enumerate(
        [
            out
            for out in alg.destinationParameterDefinitions()
            if not out.flags() & QgsProcessingParameterDefinition.Flag.FlagHidden
        ]
    ):
        if not out.name() in parameters:
            continue

        token = parameters[out.name()]

        if isinstance(out, QgsProcessingParameterRasterDestination):
            if token is None:
                QMessageBox.warning(
                    None,
                    tr("Error"),
                    tr(
                        "Seems some outputs are temporary "
                        "files. To create test you need to "
                        "redirect all algorithm outputs to "
                        "files"
                    ),
                )
                return

            try:
                dataset = gdal.Open(token, GA_ReadOnly)
            except Exception:
                QMessageBox.warning(
                    None,
                    tr("Error"),
                    tr(
                        "Seems some outputs are temporary "
                        "files. To create test you need to "
                        "redirect all algorithm outputs to "
                        "files"
                    ),
                )
                return

            dataArray = nan_to_num(dataset.ReadAsArray(0))
            strhash = hashlib.sha224(dataArray.data).hexdigest()

            results[out.name()] = {"type": "rasterhash", "hash": strhash}
        elif isinstance(
            out,
            (
                QgsProcessingParameterVectorDestination,
                QgsProcessingParameterFeatureSink,
            ),
        ):
            schema, filepath = extractSchemaPath(token)
            results[out.name()] = {"type": "vector", "name": filepath}
            if not schema:
                results[out.name()]["location"] = (
                    "[The expected result data is not in the testdata directory. Please write it to processing/tests/testdata/expected. Prefer gml files.]"
                )
        elif isinstance(out, QgsProcessingParameterFileDestination):
            schema, filepath = extractSchemaPath(token)
            results[out.name()] = {"type": "file", "name": filepath}
            if not schema:
                results[out.name()]["location"] = (
                    "[The expected result file is not in the testdata directory. Please redirect the output to processing/tests/testdata/expected.]"
                )

    definition["results"] = results
    dlg = ShowTestDialog(yaml.dump([definition], default_flow_style=False))
    dlg.exec()


def tr(string):
    return QCoreApplication.translate("TestTools", string)


class ShowTestDialog(QDialog):
    def __init__(self, s):
        QDialog.__init__(self)
        self.setModal(True)
        self.resize(600, 400)
        self.setWindowTitle(self.tr("Unit Test"))
        layout = QVBoxLayout()
        self.text = QTextEdit()
        self.text.setFontFamily("monospace")
        self.text.setEnabled(True)
        # Add two spaces in front of each text for faster copy/paste
        self.text.setText("  {}".format(s.replace("\n", "\n  ")))
        layout.addWidget(self.text)
        self.setLayout(layout)
        QMetaObject.connectSlotsByName(self)
